# MIT License
# 
# Copyright (c) 2022, Alex M. Maldonado
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Groups atoms into fragments."""

import numpy as np
from scipy.spatial.distance import cdist

def pairwise_distance(R, cutoff=1.5, num_frags=None):
    """Group atoms together into fragments based on some cutoff distance.

    This is not very dependable.

    Parameters
    ---------
    R : :obj:`numpy.ndarray`
        Cartesian coordinates of all atoms in the system in Angstroms.
    cutoff : :obj:`float`
        Atoms within this distance is group together in a fragment.
    num_frags : :obj:`int`, optional
        Known number of fragments in structure.
    
    Returns
    -------
    :obj:`numpy.ndarray`

    """
    if R.ndim != 2:
        raise ValueError('Array must have two dimensions.')

    entity_ids = np.empty((R.shape[0],))
    entity_ids[:] = np.NaN

    all_distances = cdist(R, R, 'euclidean')

    frag_idx = 0
    for i in range(len(entity_ids)):
        frag_atoms = np.argwhere(all_distances[i]<cutoff).T[0]
        # If one of the atoms are already in fragment we include these atoms.
        atoms_entity_ids = entity_ids[frag_atoms]
        if np.isnan(atoms_entity_ids).all():
            entity_ids[frag_atoms] = frag_idx
            frag_idx += 1
        else:
            prev_frag_idx = np.unique(
                atoms_entity_ids[~np.isnan(atoms_entity_ids)]
            )
            assert len(prev_frag_idx) == 1
            prev_frag_idx = prev_frag_idx[0]
            entity_ids[frag_atoms] = prev_frag_idx
    
    if num_frags is not None:
        if num_frags != frag_idx-1:
            raise ValueError(
                f'Detected {frag_idx-1} fragments instead of {num_frags}.'
            )
    
    return entity_ids




# MIT License
# 
# Copyright (c) 2022, Alex M. Maldonado
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from .extractor import extractor

class extractorXTB(extractor):
    """xTB extractor"""
    def __init__(self):
        super().__init__()
    
    @property
    def triggers(self):
        trig = (
            (lambda line: True if ('* xtb version' in line) else False, 'xtb_version'),
            (lambda line: True if ('program call               :' in line.strip()) else False, 'run_type'),
            (lambda line: True if ('number of electrons        :' in line.strip()) else False, 'n_electrons'),
            (lambda line: True if ('charge                     :' in line.strip()) else False, 'charge'),
            (lambda line: True if ('spin                       :' in line.strip()) else False, 'multiplicity'),
            (lambda line: True if ('> wall' == line.strip()) else False, 'wall_pot'),
            (lambda line: True if (':                      SETUP                      :' == line.strip()) else False, 'gfn_setup'),
            (lambda line: True if ('::                     SUMMARY                     ::' in line.strip()) else False, 'summary_energies'),
            (lambda line: True if ('|               Molecular Dynamics                |' in line.strip()) else False, 'md_setup'),
            (lambda line: True if ('A N C O P T' == line.strip('| \n') or 'L-ANC optimizer' == line.strip('| \n')) else False, 'opt_data'),
            (lambda line: True if ('average properties' == line.strip()) else False, 'md_avg_props'),
            (lambda line: True if ('thermostating problem' == line.strip()) else False, 'thermostat_prob'),
            (lambda line: True if ('normal termination of xtb' == line.strip()) else False, 'success'),
        )
        return trig
    
    def xtb_version(self, f, line):
        """Version of xtb.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            * xtb version 6.3.3 (5b13467) compiled by 'ehlert@majestix' on 2020-09-17
        """
        line_split = line.split()
        version = line_split[3]
        self.parsed_info['runtime_info']['prov_version'] = version
        line = next(f)
    
    def run_type(self, f, line):
        """Calculation type (e.g., opt, sp, grad).

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            program call               : xtb 5h2o.example.xyz --scc --gfn 2 --charge 0
        """
        if '--scc' in line:
            driver = 'energy'
        elif '--opt' in line:
            driver = 'optimization'
        elif '--grad' in line:
            driver = 'gradient'
        elif '--ohess' in line or '--hess' in line:
            driver = 'frequency'
        elif '--omd' in line or '--md' in line:
            driver = 'molecular_dynamics'
        self.parsed_info['runtime_info']['calc_driver'] = driver
        line = next(f)
    
    def n_electrons(self, f, line):
        """Total number of electrons.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            number of electrons        :                  1688
        """
        n_ele = line.strip(' :').split()[-1]
        self.parsed_info['system_info']['n_ele'] = int(n_ele)

    def charge(self, f, line):
        """Overall system charge.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            charge                     :                     0
        """
        _, _, charge = line.strip(' :').split()
        self.parsed_info['system_info']['charge'] = int(charge)
    
    def multiplicity(self, f, line):
        """Overall system multiplicity.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            spin                       :                   0.0
        """
        _, _, spin = line.strip().split()
        mult = 2 * float(spin) + 1
        self.parsed_info['system_info']['mult'] = int(mult)
    
    def wall_pot(self, f, line):
        """Wall potentials.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            > wall
            -> potential = logfermi
            -> sphere: 23.62158, all
            --> 1: 23.62158
            --> 2: all
        """
        if 'wall_potential' in self.parsed_info.keys():
            pots = self.parsed_info['runtime_info']['wall_potential']
        else:
            pots = []
        idx_start = len(pots)
        
        line = next(f)
        pot_type = line.split()[3]

        line = next(f)
        while '$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$' not in line:
            if '-> sphere:' == line[:10]:
                shape_type = line.split(':')[0].split(' ')[1]
                line = next(f)
                line = next(f)
                atoms_constrained = ''
                while 'spherical wallpotenial' != line[:22]:
                    line_split = line.strip().split()
                    if 'all' == line_split[-1]:
                        atoms_constrained = 'all'
                    else:
                        atoms_constrained += f'{line_split[-1]},'
                    line = next(f)
                if atoms_constrained[-1] == ',':
                    atoms_constrained = atoms_constrained[:-1]
                sphere_radius = float(line.split()[-2])
                pots.append(
                    {
                        'shape': shape_type,
                        'sphere_radius': sphere_radius,
                        'atoms_constrained': atoms_constrained,
                    }
                )
            elif '-> ellipsoid:' == line[:13]:
                pass

            if '-> temp =' == line[:9]:
                temp = float(line.split()[-1])
                line = next(f)
                beta = float(line.split()[-1])
                for i in range(idx_start, len(pots)):
                    pots[i]['logfermi_temp'] = temp
                    pots[i]['logfermi_beta'] = beta
                
            line = next(f)

        self.parsed_info['runtime_info']['wall_potential'] = pots
    
    def gfn_setup(self, f, line):
        """GFN calculation setup.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            ...................................................
            :                      SETUP                      :
            :.................................................:
        """
        line = self.skip_lines(f, 2)
        while '..............................' not in line.strip():
            
            # :  # basis functions                1446          :
            if ' # basis functions' in line:
                self.parsed_info['runtime_info']['basis_n_func'] = int(line.strip(' :').split()[-2])
            
            # :  Hamiltonian                  GFN2-xTB          :
            if ':  Hamiltonian ' in line:
                self.parsed_info['runtime_info']['hamiltonian'] = line.strip(' :').split()[1]
            
            line = next(f)
    
    def summary_energies(self, f, line):
        """Extracts energies listed in SUMMARY box.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            :::::::::::::::::::::::::::::::::::::::::::::::::::::
            ::                     SUMMARY                     ::
            :::::::::::::::::::::::::::::::::::::::::::::::::::::
            :: total energy             -25.404338997931 Eh    ::
            :: gradient norm              0.048763527905 Eh/a0 ::
            :: HOMO-LUMO gap             12.772900340050 eV    ::
            ::.................................................::
            :: SCC energy               -25.580127846040 Eh    ::
            :: -> isotropic ES            0.153316435656 Eh    ::
            :: -> anisotropic ES         -0.020934020006 Eh    ::
            :: -> anisotropic XC         -0.003785806991 Eh    ::
            :: -> dispersion             -0.004190013801 Eh    ::
            :: repulsion energy           0.175746000529 Eh    ::
            :: add. restraining           0.000000000000 Eh    ::
            :: total charge              -0.000000000000 e     ::
            :::::::::::::::::::::::::::::::::::::::::::::::::::::
        """
        if self.parsed_info['runtime_info']['calc_driver'] != 'molecular_dynamics':
            if 'energy_scf' not in self.parsed_info['outputs'].keys():
                self.parsed_info['outputs']['energy_scf'] = []
            
            if 'energy_nuc_repul' not in self.parsed_info['outputs'].keys():
                self.parsed_info['outputs']['energy_nuc_repul'] = []
            
            while '' != line.strip():
                if 'total energy' in line:
                    line_split = line.split()
                    energy_scf = float(line_split[3])
                
                if 'repulsion energy' in line:
                    line_split = line.split()
                    energy_nuc_repul = float(line_split[3])

                line = next(f)

            self.parsed_info['outputs']['energy_scf'].append(
                energy_scf
            )
            self.parsed_info['outputs']['energy_nuc_repul'].append(
                energy_nuc_repul
            )
        # For MD simulations.
        else:
            if 'energy_pot' not in self.parsed_info['outputs'].keys():
                self.parsed_info['outputs']['energy_pot'] = []
            
            while '' != line.strip():
                if 'total energy' in line:
                    line_split = line.split()
                    energy_pot = float(line_split[3])
                    self.parsed_info['outputs']['energy_pot'].append(
                        energy_pot
                    )

                line = next(f)
    
    def md_setup(self, f, line):
        """Extracts MD 

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

                     ------------------------------------------------- 
                    |               Molecular Dynamics                |
                     ------------------------------------------------- 
            trajectories on xtb.trj or xtb.trj.<n>
                                
            MD time /ps        :    5.00
            dt /fs             :    1.00
            SCC accuracy       :    1.00
            temperature /K     :  500.00
            max steps          :  5000
        """
        while 'MD time /ps' not in line:
            line = next(f)
        
        #  MD time /ps        :    5.00
        self.parsed_info['runtime_info']['md_duration'] = float(line.split()[4])

        line = next(f)
        #  dt /fs             :    1.00
        self.parsed_info['runtime_info']['t_step'] = float(line.split()[3])

        line = next(f)
        #  SCC accuracy       :    1.00
        self.parsed_info['runtime_info']['xtb_scc_accuracy'] = float(line.split()[3])

        line = next(f)
        #  temperature /K     :  500.00
        self.parsed_info['runtime_info']['thermostat_temp'] = float(line.split()[3])

        while 'dumpstep(trj) /fs' not in line:
            line = next(f)
        
        #  dumpstep(trj) /fs  :    5.00     5
        self.parsed_info['runtime_info']['md_steps_dump_traj'] = float(line.split()[4])

        while 'time (ps)    <Epot>' not in line:

            #  H atoms mass (amu) :     1
            if 'H atoms mass (amu) :' == line.strip()[:20]:
                self.parsed_info['runtime_info']['mass_h'] = float(line.split()[-1])
            
            # TODO: Parse shake on information.
            if 'SHAKE off' == line.strip():
                self.parsed_info['runtime_info']['algo_shake'] = False
            
            if 'Berendsen THERMOSTAT on' == line.strip():
                self.parsed_info['runtime_info']['thermostat_type'] = 'Berendsen'
            
            if 'RESTART' == line.strip():
                self.parsed_info['runtime_info']['md_restarted'] = True
            
            line = next(f)
        
        if 'md_restarted' not in self.parsed_info['runtime_info'].keys():
            self.parsed_info['runtime_info']['md_restarted'] = False
        
    
    def opt_data(self, f, line):
        """All information during optimization routine. Incldues setup, energy.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

             ----------------------------------------------------------- 
            |                   =====================                   |
            |                        A N C O P T                        |
            |                   =====================                   |
            |               Approximate Normal Coordinate               |
            |                Rational Function Optimizer                |
             ----------------------------------------------------------- 
        
        .. code-block:: text

             ----------------------------------------------------------- 
            |                       L-ANC optimizer                     |
             ----------------------------------------------------------- 
        """
        
        # The following lines signal the termination of the optimization routine.
        # *** GEOMETRY OPTIMIZATION CONVERGED AFTER 73 ITERATIONS ***
        # *** FAILED TO CONVERGE GEOMETRY OPTIMIZATION IN 200 CYCLES ***
        # Note that some lines have '(*******%)' in them, so we add spaces around
        # The asterisks to avoid triggering early.
        while ' *** ' not in line and 'GEOMETRY OPTIMIZATION' not in line:
            if '* total energy  :' in line:
                line_split = line.split()
                energy_scf = float(line_split[4])
                self.parsed_info['outputs']['energy_scf'].append(
                    energy_scf
                )
            line = next(f)
    
    def md_avg_props(self, f, line):
        """Average MD properties.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            average properties 
            Epot               :  -991.673089209627     
            Epot (accurate SCC):  -991.771829504958     
            Ekin               :   1.79432794141482     
            Etot               :  -989.878761268212     
            T                  :   522.456878068899  
        """
        while 'T                  :' not in line:
            if 'Etot               :' in line:
                self.parsed_info['outputs']['avg_energy_tot'] = float(line.split()[-1])
            line = next(f)
        
        self.parsed_info['outputs']['avg_temp'] = float(line.split()[-1])
        
    def thermostat_prob(self, f, line):
        """If there was an issue with the thermostat.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            thermostating problem
        """
        self.parsed_info['runtime_info']['success'] = False
    
    def success(self, f, line):
        """If the calculation is successful.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            normal termination of xtb
        """
        if 'success' not in self.parsed_info['runtime_info'].keys():
            self.parsed_info['runtime_info']['success'] = True
    


# MIT License
# 
# Copyright (c) 2022, Alex M. Maldonado
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from .writer import reptarWriter
from ..utils import z_to_element

class pdbWriter(reptarWriter):
    """Writes PDB file from reptar group.
    """

    def __init__(self):
        super().__init__()
    
    def write(
        self, data, group_key, atom_type='HETATM', file_name=None, save_dir=None,
        R_limits=None
    ):
        """Parses trajectory file and extracts information.

        Parameters
        ----------
        data : ``reptar.data``
            A reptar data object.
        group_key : :obj:`str`
            Key to the desired group.
        atom_type : :obj:`str`, optional
            The PDB atom type to be used. Should almost always be ``HETATM``.
        file_name : :obj:`str`, optional
            Name of the PDB file. Defaults to ``'structure'``.
        save_dir : :obj:`str`, optional
            Directory to save the PDB file. Defaults to the current directory.
        R_limits : :obj:`tuple` (:obj:`int`), optional
            Slicing limits specifying start and stop index. If ``None`` is used
            it does not change the beginning or end of the array. For example,
            ``(None, 100)`` would only print the first 100 structures.
        """
        R = data.get(f'{group_key}/geometry')
        if R.ndim == 2:
            R = np.array([R])
        if R_limits is not None:
            R = R[slice(*R_limits)]
        Z = data.get(f'{group_key}/atomic_numbers')
        atom_labels = [z_to_element[i] for i in Z]
        entity_ids = data.get(f'{group_key}/entity_ids')
        comp_ids = data.get(f'{group_key}/comp_ids')
        
        if file_name is None:
            file_name = 'structure'
        if save_dir is None:
            save_dir = '.'
        else:
            if save_dir[-1] == '/':
                save_dir = save_dir[:-1]

        num_structures = len(R)
        num_atoms = len(Z)
        with open(f'{save_dir}/{file_name}.pdb', 'w') as f:
            f.write(file_name+'\n')
            f.write(f'{num_atoms}\n')  
            for i_structure in range(num_structures):
                if i_structure != 0:
                    f.write('MODEL\n')
                for i_atom in range(num_atoms):
                    entity_id = entity_ids[i_atom]
                    comp_id = comp_ids[entity_id]  # TODO: Handle component ids that are longer than three letters.
                    atom_label = atom_labels[i_atom]
                    coords = R[i_structure][i_atom]
        
                    f.write(
                        '{:6s}{:5d} {:^4s}{:1s}{:3s} {:1s}{:4d}{:1s}   {:8.3f}{:8.3f}{:8.3f}{:6.2f}{:6.2s}          {:>2s}{:2s}\n'.format(
                            str(atom_type), i_atom+1, str(atom_label), '',
                            str(comp_id), 'A', entity_id+1, '',
                            coords[0], coords[1], coords[2],
                            1.00, '', str(atom_label), ''
                        )
                    )
                if num_structures > 1:
                    f.write('ENDMDL\n')
                
    
    


# MIT License
# 
# Copyright (c) 2022, Alex M. Maldonado
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from .parser import parser
from ..extractors import extractorORCA
import cclib

class parserORCA(parser):
    """Custom parser for ORCA output files.

    Parameters
    ----------
    out_path : :obj:`str`
        Path to the main log file generated by the package.
    geom_path : :obj:`str`, optional
        Path to a file containing a single geometry.
    traj_path : :obj:`str`, optional
        Path to a trajectory file from a geometry optimization, MD
        simulation, etc.
    extractors : :obj:`list`, optional
        Additional extractors for the parser to use.
    """

    def __init__(self, out_path, geom_path=None, traj_path=None, extractors=None):
        self.package = 'orca'
        if extractors is None:
            extractors = []
        extractors.insert(0, extractorORCA())
        super().__init__(out_path, extractors)
        # TODO: Handle geom and traj paths
    
    def parse(self):
        """Parses output file and extracts information.
        """
        # cclib parsed information.
        try:
            self.cclib_data = cclib.io.ccread(self.out_path)
            self.map_cclib_data()
        except Exception as e:
            raise e

        # Extract information.
        self.extract_data_out()

        # Postprocessing
        self.after_parse()
        # TODO: Figure out how to get driver.

        return self.parsed_info

    def after_parse(self):
        """Checks to perform after parsing output file.
        """
        if 'scf_grid_level_final' not in self.parsed_info['runtime_info'].keys():
            self.parsed_info['runtime_info']['scf_grid_level_final'] = \
                self.parsed_info['runtime_info']['scf_grid_level']
        
        if 'conv_val_geo_energy' in self.parsed_info['runtime_info'].keys():
            self.parsed_info['runtime_info']['conv_val_geo_energy'].insert(0, 0.0)
    



# MIT License
# 
# Copyright (c) 2022, Alex M. Maldonado
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
from xtb.interface import Calculator, Param

def xtb_engrad(
    Z, R, R_idxs, charge, mult, calc_acc=0.1, params=Param.GFN2xTB
):
    """Ray remote function for computing total electronic energy and atomic
    gradients using xtb.

    Parameters
    ----------
    Z : ``ray.ObjectRef`` of :obj:`numpy.ndarray`
        Atomic numbers of the atoms with repsect to ``R``.
    R : ``ray.ObjectRef`` of :obj:`numpy.ndarray`
        Cartesian coordinates of all structures in group. This includes
        unused structures.
    R_idxs : :obj:`numpy.ndarray`
        Indices of the structures from ``R`` to compute energies and gradients
        for.
    charge : :obj:`int`
        Total molecular charge.
    mult : :obj:`int`
        Total multiplicity.
    calc_acc : :obj:`str`, optional
        Numerical accuracy for calculation. For more information, visit the
        `documentation <https://xtb-python.readthedocs.io/en/latest\
        /general-api.html#xtb.interface.Calculator.set_accuracy>`_.
        Defaults to ``0.1``.
    params, optional
        xTB parameters. Defaults to ``Param.GFN2xTB``.
    
    Returns
    -------
    :obj:`numpy.ndarray`
        ``R_idxs``
    :obj:`numpy.ndarray`
        Total electronic energy of computed structures in the same order as
        ``R_idxs``. Units of Hartree.
    :obj:`numpy.ndarray`
        Atomic gradients of computed structures in the same order as ``R_idxs``.
        Units of Hartree/Angstrom.
    """
    n_upair_ele = int(mult - 1)
    R = R[R_idxs]
    G = np.zeros(R.shape)
    E = np.zeros(R.shape[0])
    for i in range(len(R)):
        calc = Calculator(Param.GFN2xTB, Z, R[i], charge, n_upair_ele)
        calc.set_accuracy(calc_acc)
        res = calc.singlepoint()
        g = res.get_gradient()
        g /= 0.52917721067  # psi4.constants.bohr2angstroms
        G[i] = g
        E[i] = res.get_energy()
    return R_idxs, E, G



__author__ = """Alex M. Maldonado"""

from . import _version
__version__ = _version.get_versions()['version']

from . import utils
from . import extractors
from . import parsers
from .creator import creator
from .data import data
from . import sampler
from . import writers
from . import fragment

__all__ = [
    'utils', 'extractors', 'parsers', 'creator', 'data', 'sampler', 'writers',
    'fragment'
]



# MIT License
# 
# Copyright (c) 2022, Alex M. Maldonado
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import shutil
import numpy as np
from .parsers import parserORCA, parserXTB
from .utils import get_md5
from .data import data
from pkg_resources import resource_stream
import yaml

from reptar import _version
__version__ = _version.get_versions()['version']

defs_reserved = [
    'base', 'md', 'molprop', 'pes', 'qc', 'sampl', 'solv', 'xtb'
]

def identify_parser(out_path):
    """Identifies the correct parser depending on some trigger.
    Only supported packaged should be included in `triggers`.

    Parameters
    ----------
    out_path : :obj:`str`
        Path to output file.
    
    Returns
    -------
    :obj:`reptar.creator`
        One of the supported creator classes.
    """
    with open(out_path, 'r') as f:
        for line in f:
            for parser, phrases, do_break in triggers:
                if all([line.lower().find(p.lower()) >= 0 for p in phrases]):
                    filetype = parser
                    if do_break:
                        return filetype

# Triggers to identify output files.
triggers = [
    (parserORCA, ["O   R   C   A"], True),
    (parserXTB, ["x T B"], True)
]

class creator:
    """Create groups from computational chemistry data.

    Parameters
    ----------
    data : ``reptar.data``, optional
        An initialized data object.
    """

    def __init__(self, data=None):
        if data is not None:
            self.data = data
    
    def load(
        self, file_path, mode='r', allow_remove=False, plugins=None
    ):
        """Load a data file for creating/adding information.

        Parameters
        ----------
        file_path : :obj:`str`
            Path to a file supported by reptar. If it does not exist, then one will
            be created if possible.
        mode : :obj:`str`, optional
            A file mode string that defines the read/write behavior. Defaults to
            ``'r'``.
        allow_remove : :obj:`bool`, optional
            Allow the removal of exdir groups in ``w`` operation. Defaults to
            ``False``.
        plugins : :obj:`list`, optional
            A list of instantiated exdir plugins. Defaults to ``None``.
        """
        self.data = data(
            file_path, mode=mode, allow_remove=allow_remove, plugins=plugins
        )

    def parse_output(
        self, out_path, geom_path=None, traj_path=None, extractors=None
    ):
        """Parse output file using cclib and custom parser.

        Parameters
        ----------
        out_path : :obj:`str`
            Path to computational chemistry output file.
        geom_path : :obj:`str`, optional
            Path to a file containing a single geometry.
        traj_path : :obj:`str`, optional
            Path to a trajectory file from a geometry optimization, MD
            simulation, etc.
        extractors : :obj:`list`, optional
            Additional extractors for the parser to use.
        """
        self.out_path = os.path.abspath(out_path)
        if geom_path is not None:
            geom_path = os.path.abspath(geom_path)
        if traj_path is not None:
            traj_path = os.path.abspath(traj_path)
        
        self.geom_path = geom_path
        self.traj_path = traj_path

        packageParser = identify_parser(self.out_path)
        self.parser = packageParser(
            self.out_path, geom_path=geom_path, traj_path=traj_path,
            extractors=extractors
        )
        self.parsed_info = self.parser.parse()
    
    def _create_extras_xtb(self, group_key):
        """Extra actions for groups created from xtb calculations.

        Parameters
        ----------
        group_key : :obj:`str`
            Key to the desired new group (including parent).
        """
        out_dir = os.path.dirname(self.out_path)
        group = self.data.get(group_key)
        
        if self.data.ftype == 'exdir':
            md_restart_path = f'{out_dir}/mdrestart'
            if os.path.exists(md_restart_path):
                raw = group.require_raw('restart_files')
                raw_path = os.path.join(raw.root_directory, raw.relative_path)
                shutil.copy(
                    md_restart_path, os.path.join(raw_path, 'mdrestart')
                )
            xtb_restart_path = f'{out_dir}/xtbrestart'
            if os.path.exists(xtb_restart_path):
                raw = group.require_raw('restart_files')
                raw_path = os.path.join(raw.root_directory, raw.relative_path)
                shutil.copy(
                    xtb_restart_path, os.path.join(raw_path, 'xtbrestart')
                )
    
    def group(
        self, group_key, out_path=None, geom_path=None, traj_path=None,
        extractors=None
    ):
        """Create a group using any data reptar can find from provided paths.
        
        Parameters
        ----------
        group_key : :obj:`str`
            Key to the desired new group (including parent).
        out_path : :obj:`str`, optional
            Path to the main log file generated by the package.
        geom_path : :obj:`str`, optional
            Path to a file containing a single geometry.
        traj_path : :obj:`str`, optional
            Path to a trajectory file from a geometry optimization, MD
            simulation, etc.
        extractors : :obj:`list`, optional
            Additional extractors for the parser to use.
        
        Notes
        -----
        If both ``geom_path`` and ``traj_path`` are provided, it is assumed that
        ``geom_path`` provides an initial geometry not included in
        ``traj_path``.
        """
        assert hasattr(self, 'data')
        self.parse_output(
            out_path, geom_path=geom_path, traj_path=traj_path,
            extractors=extractors
        )
        parsed_info = self.parsed_info

        if self.data.ftype == 'exdir':
            self.data.init_group(group_key)
        
        # Loop through each category of data.
        for cat_key in parsed_info.keys():
            for data_key in parsed_info[cat_key].keys():
                data = parsed_info[cat_key][data_key]
                self.data.add(f'{group_key}/{data_key}', data)
        
        # MD5 stuff
        md5 = get_md5(self.data, group_key)
        self.data.add(f'{group_key}/md5', md5)
        
        try:
            md5_arrays = get_md5(self.data, group_key, only_arrays=True)
            self.data.add(f'{group_key}/md5_arrays', md5_arrays)
        except Exception:
            pass

        try:
            md5_structures = get_md5(self.data, group_key, only_structures=True)
            self.data.add(f'{group_key}/md5_structures', md5_structures)
        except Exception:
            pass

        # Extra stuff to do depending on package.
        if self.parser.package == 'xtb':
            self._create_extras_xtb(group_key)
        
        # Adding version
        self.data.add(f'{group_key}/reptar_version', __version__)

        return self.data
    
    def ids(self, group_key, entity_ids, comp_ids):
        """Add ``entity_ids`` and ``comp_ids`` to a group.

        Parameters
        ----------
        group_key : :obj:`str`
            Key to the group to add IDs to (including parent).
        entity_ids : :obj:`numpy.ndarray`
            A uniquely identifying integer specifying what atoms belong to which
            entities. Entities can be a related set of atoms, molecules, or
            functional group.
        comp_ids : :obj:`numpy.ndarray`
            Relates ``entity_id`` to a fragment label for chemical components
            or species. Labels could be ``WAT`` or ``h2o`` for water, ``MeOH``
            for methanol, ``bz`` for benzene, etc. There are no standardized
            labels for species. The index of the label is the respective
            ``entity_id``.
        """
        if isinstance(comp_ids, list):
            comp_ids = np.array(comp_ids)
        
        self.data.add(f'{group_key}/entity_ids', entity_ids)
        self.data.add(f'{group_key}/comp_ids', comp_ids)

        comp_ids_num = {}
        unique_comp_ids, comp_ids_freq = np.unique(comp_ids, return_counts=True)
        for comp_id, num in zip(unique_comp_ids, comp_ids_freq):
            comp_ids_num[str(comp_id)] = int(num)
        
        self.data.add(f'{group_key}/comp_ids_num', comp_ids_num)
        return self.data.get(group_key)
    
    # TODO: Update ways we get data here.
    def definitions(self, definitions=None):
        """Add definitions of data to the file.

        The base definitions are included by default. There should be no overlap
        of any definitions. If there is, precedence is given in order of
        ``definitions`` with base always being first.

        Parameters
        ----------
        definitions : :obj:`list` [:obj:`str`], optional
            Paths to data definition YAML files to congregate. Only the name
            (not a path) is required for ones provided by reptar.
        
        Notes
        -----
        Reserved definition YAML files: ``base``, ``md``, ``molprop``, ``pes``,
        ``qc``, ``sampl``, ``solv``, ``xtb``.
        """        
        defs = {}

        if definitions is None:
            definitions = ['base']
        else:
            definitions.insert(0, 'base')
        
        for def_path in definitions:
            if def_path in defs_reserved:
                stream = resource_stream(
                    'reptar.definitions', f'{def_path}.yaml'
                )
            else:
                stream = open(def_path)
            def_add = yaml.safe_load(stream)
            
            # Add all definitions.
            for key_cat in def_add.keys():
                if key_cat not in defs.keys():
                    defs[key_cat] = {}
                for key_def in def_add[key_cat].keys():
                    if key_def not in defs[key_cat].keys():
                        defs[key_cat][key_def] = def_add[key_cat][key_def]
                    else:
                        pass
        
        self.data.add('definitions', defs)



# MIT License
# 
# Copyright (c) 2022, Alex M. Maldonado
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from abc import ABC, abstractmethod

class extractor(ABC):
    """Base class for extracting information from files.
    """
    def __init__(self):
        self.parsed_info = {
            'system_info': {},
            'runtime_info': {},
            'outputs': {}
        }

    @property
    @abstractmethod
    def triggers(self):
        """A collection of triggers that activate the corresponding extractor.

        The trigger is a lambda function that returns True or False depending
        on the criteria and the name of the extractor method.

        :type: :obj:`tuple`
        """
        pass
    
    @property
    def parsed_info(self):
        """Information parsed from files. Contains the following keys.

        ``system_info``
            Information specifying the system prior to any computation. Such
            as the initial cartesian coordinates, total system charge and
            multiplicity, etc.
        
        ``runtime_info``
            Contains information about setting up the job/calculation or running
            the job. Defining convergence criteria, parameters, etc.
        
        ``outputs``
            Results, requested or not, of the job. For example, SCF
            cycle values, optimized coordinates, trajectory, number of
            electrons, generated structures, etc.

        :type: :obj:`dict`
        """
        return self._parsed_info
    
    @parsed_info.setter
    def parsed_info(self, value):
        self._parsed_info = value

    @parsed_info.deleter
    def parsed_info(self, value):
        del self._parsed_info
    
    def skip_lines(self, f, n):
        """Skip a number of lines.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        n : :obj:`int`
            Number of lines to skip
        
        Returns
        -------
        :obj:`str`
            Line of file.
        """
        for i in range(n):
            line = next(f)
        return line
    


# MIT License
# 
# Copyright (c) 2022, Alex M. Maldonado
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from .extractor import extractor

class extractorORCA(extractor):
    """ORCA extractor"""
    def __init__(self):
        super().__init__()
    
    @property
    def triggers(self):
        trig = (
            (lambda line: True if ('Your calculation utilizes the auxiliary basis: ' in line.strip()) else False, 'aux_basis'),
            (lambda line: True if ('DFT GRID GENERATION' == line.strip() or 'Setting up the final grid:' == line.strip()) else False, 'grid_info'),
            (lambda line: True if ('CPCM SOLVATION MODEL' == line.strip()) else False, 'implicit_solvent'),
            (lambda line: True if ('TOTAL SCF ENERGY' == line.strip()) else False, 'scf_energies'),
            (lambda line: True if ('UHF SPIN CONTAMINATION' == line.strip()) else False, 'uhf_spin_contamination'),
            (lambda line: True if ('ORCA  MP2' == line.strip()) else False, 'mp_properties'),
            (lambda line: True if ('Wavefunction type' == line.strip()) else False, 'cc_method'),
            (lambda line: True if ('HF COUPLED CLUSTER ITERATIONS' == line.strip()[1:] or 'OPEN-SHELL COUPLED CLUSTER ITERATIONS' == line.strip()) else False, 'cc_properties'),
            (lambda line: True if ('SCF SETTINGS' == line.strip()) else False, 'scf_info'),
            (lambda line: True if ('MULLIKEN ATOMIC CHARGES' == line.strip()) else False, 'mulliken_charges'),
            (lambda line: True if ('LOEWDIN ATOMIC CHARGES' == line.strip()) else False, 'loewdin_charges'),
            (lambda line: True if ('DIPOLE MOMENT' == line.strip()) else False, 'dipole'),
            (lambda line: True if ('Geometry convergence' in line and '----------------------' in line) else False, 'geo_conv'),
            (lambda line: True if ('VIBRATIONAL FREQUENCIES' == line[:23]) else False, 'frequencies'),
            (lambda line: True if ('NORMAL MODES' == line[:12]) else False, 'normal_modes'),
            (lambda line: True if ('Temperature         ...' in line) else False, 'thermo_temp'),
            (lambda line: True if ('Zero point energy                ...' in line) else False, 'zpve'),
            (lambda line: True if ('Total thermal correction' in line) else False, 'thermal_corrections'),
            (lambda line: True if ('Thermal Enthalpy correction       ...' in line) else False, 'enthalpic_corr'),
            (lambda line: True if ('Final entropy term                ...' in line) else False, 'entropic_corr'),
        )
        return trig

    def aux_basis(self, f, line):
        """Information about auxiliary basis sets used in the calculation.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            ----- AuxJ basis set information -----
            Your calculation utilizes the auxiliary basis: def2/J
        """
        line_split = line.strip().split(':')
        self.parsed_info['runtime_info']['basis_aux'] = line_split[-1].strip()

    def grid_info(self, f, line):
        """DFT integration grid information.

        All DFT calculations are performed with numerical integration, which
        means a integration grid must be specified. This is typically handeled
        with the Grid keyword. Furthermore, ORCA defaults to a multigrid
        approach where one grid is used for the SCF cycle, and a different
        (typically larger) grid is used for the final energy evaluation.

        After testing different keyword combinations, the Grid keywords are not
        always consistent with the final grid. Thus, we are going to directly
        parse the grid information from the output file instead of depending
        on the keywords.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            -------------------
            DFT GRID GENERATION
            -------------------
        
        Here are the main parameters specifying the ORCA default grid levels.
        Default SCF grid is 2.

        +--------+---------------+----------+
        |  Grid  |  AngularGrid  |  IntAcc  |
        +========+===============+==========+
        |  1     |  Lebedev-50   |  4.34    |
        |  2     |  Lebedev-110  |  4.34    |
        |  3     |  Lebedev-194  |  4.34    |
        |  4     |  Lebedev-302  |  4.67    |
        |  5     |  Lebedev-434  |  5.01    |
        |  6     |  Lebedev-590  |  5.34    |
        |  7     |  Lebedev-770  |  5.67    |
        +--------+---------------+----------+
        """
        lebedev_to_level = {
            '50': 1, '110': 2, '194': 3, '302': 4, '434': 5, '590': 6, '770': 7
        }
        
        # SCF Cycle Grid
        if 'DFT GRID GENERATION' == line.strip():
            while 'Angular Grid (max. acc.)' not in line:
                line = next(f)
            lebedev_num = line.strip().split('-')[-1]
            self.parsed_info['runtime_info']['scf_grid_level'] = lebedev_to_level[lebedev_num]
        elif 'Setting up the final grid:' == line.strip():
            while 'Angular Grid (max. acc.)' not in line:
                line = next(f)
            lebedev_num = line.strip().split('-')[-1]
            self.parsed_info['runtime_info']['scf_grid_level_final'] = lebedev_to_level[lebedev_num]

    def implicit_solvent(self, f, line):
        """Implicit solvent properties.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            --------------------
            CPCM SOLVATION MODEL
            --------------------
        """
        while 'Overall time for CPCM initialization' != line.strip()[:36]:
            # Solvent:                                     ... ACETONITRILE
            if 'Solvent:' == line[:8]:
                self.parsed_info['runtime_info']['solv_implicit_name'] = line.split()[2]
            
            line = next(f)

    
    def scf_energies(self, f, line):
        """The nuclear repulsion, one- and two-electron energy, and
        exchange-correlation energy after a SCF cycle.

        This is called directly after the ``'TOTAL SCF ENERGY'`` trigger, and 
        will terminate once the ``'SCF CONVERGENCE'`` trigger is reached.

        Instead of returning the energies themselves, we handle the creation and
        modification of ``scf_info`` here so any missing information (such as
        ``'energy_scf_xc'`` in MP2 calculations) is not an issue.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            ----------------
            TOTAL SCF ENERGY
            ----------------
        """
        while 'SCF CONVERGENCE' != line.strip():
            # Nuclear Repulsion  :  135.87324654 Eh    3697.29901 eV
            if 'Nuclear Repulsion  :' in line:
                if 'energy_nuc_repul' not in self.parsed_info['outputs'].keys():
                    self.parsed_info['outputs']['energy_nuc_repul'] = []
                self.parsed_info['outputs']['energy_nuc_repul'].append(
                    float(line.split()[3])
                )

            # One Electron Energy: -674.26034691 Eh  -18347.55681 eV
            if 'One Electron Energy:' in line:
                if 'energy_scf_one_ele' not in self.parsed_info['outputs'].keys():
                    self.parsed_info['outputs']['energy_scf_one_ele'] = []
                self.parsed_info['outputs']['energy_scf_one_ele'].append(
                    float(line.split()[3])
                )

            # Two Electron Energy:  245.90403408 Eh    6691.38895 eV
            if 'Two Electron Energy:' in line:
                if 'energy_scf_two_ele' not in self.parsed_info['outputs'].keys():
                    self.parsed_info['outputs']['energy_scf_two_ele'] = []
                self.parsed_info['outputs']['energy_scf_two_ele'].append(
                    float(line.split()[3])
                )

            # E(XC)              :      -26.170411238000 Eh 
            if 'E(XC) ' in line:
                if 'energy_scf_xc' not in self.parsed_info['outputs'].keys():
                    self.parsed_info['outputs']['energy_scf_xc'] = []
                self.parsed_info['outputs']['energy_scf_xc'].append(
                    float(line.split()[2])
                )

            line = next(f)

    
    def uhf_spin_contamination(self, f, line):
        """Spin contamination from unrestricted Hartree-Fock calculations.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            ----------------------
            UHF SPIN CONTAMINATION
            ----------------------
            
            Expectation value of <S**2>     :     0.750016
            Ideal value S*(S+1) for S=0.5   :     0.750000
            Deviation                       :     0.000016
        """

        for _ in range(0, 3):
            line = next(f)
        if line.strip() == 'Warning: in a DFT calculation there is little theoretical justification to':
            for _ in range(0, 4):
                line = next(f)
        
        if 'spin_sqrd_uhf_ideal' not in self.parsed_info['outputs'].keys():
            self.parsed_info['outputs']['spin_sqrd_uhf_ideal'] = []
        if 'spin_sqrd_uhf_calc' not in self.parsed_info['outputs'].keys():
            self.parsed_info['outputs']['spin_sqrd_uhf_calc'] = []

        # Expectation value of <S**2>     :     0.750016
        self.parsed_info['outputs']['spin_sqrd_uhf_calc'].append(
            float(line.split()[5])
        )
        line = next(f)
        # Ideal value S*(S+1) for S=0.5   :     0.750000
        self.parsed_info['outputs']['spin_sqrd_uhf_ideal'].append(
            float(line.split()[6])
        )

    
    def mp_properties(self, f, line):
        """Moller-Plesset calculation properties.

        This is called directly after the ``'ORCA  MP2 '`` trigger, and 
        will terminate once the ``'ORCA property calculations'`` trigger is reached.

        Instead of returning the energies themselves, we handle the creation and
        modification of ``mp_info`` here so any missing information is not an
        issue.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            ----------------------------------------------------------
                                    ORCA  MP2 
            ----------------------------------------------------------
            
            Expectation value of <S**2>     :     0.750016
            Ideal value S*(S+1) for S=0.5   :     0.750000
            Deviation   
        """
        while '-     ORCA property calculations      *' != line.strip():
            # Freezing NCore=10 chemical core electrons
            if 'Freezing NCore=' == line.strip()[:15]:
                # if 'ele_frozen' not in self.data['keywords'].keys():
                #     self.data['outputs']['ele_frozen'] = []
                ele_frozen = int(line.split()[1][6:])
                self.parsed_info['outputs']['ele_frozen'] = ele_frozen

            #  MP2 CORRELATION ENERGY   :     -3.132364939 Eh
            if 'MP2 CORRELATION ENERGY' in line:
                if 'energy_correl_mp2' not in self.parsed_info['outputs'].keys():
                    self.parsed_info['outputs']['energy_correl_mp2'] = []
                if 'RI-MP2' in line:
                    index = 3
                else:
                    index = 4
                self.parsed_info['outputs']['energy_correl_mp2'].append(
                    float(line.split()[index])
                )
                break
            
            line = next(f)

    
    def cc_method(self, f, line):
        """Coupled cluster method information.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            Wavefunction type
            -----------------
            Correlation treatment                      ...      CCSD     
            Single excitations                         ... ON 
        """
        while 'Number of correlated electrons' != line.strip()[:30]:
            # Frozen core treatment                      ... chemical core (0 el)
            # or
            # Frozen core treatment                      ... NO frozen core
            if 'Frozen core treatment' == line.strip()[:21]:
                if 'NO frozen core' == line.strip()[-14:]:
                    # self.parsed_info['runtime_info']['keywords']['frozen_core'] = False
                    pass
                else:
                    ele_frozen = int(line.split()[6][1:])
                    if ele_frozen == 0:
                        # self.parsed_info['runtime_info']['keywords']['frozen_core'] = False
                        pass
                    else:
                        # self.parsed_info['runtime_info']['keywords']['frozen_core'] = True
                        self.parsed_info['runtime_info']['ele_frozen'] = ele_frozen
            
            line = next(f)

    
    def cc_properties(self, f, line):
        """Coupled cluster properties.

        This is called directly after the ``'COUPLED CLUSTER ITERATIONS'``
        trigger, and will terminate once the ``'ORCA POPULATION ANALYSIS'``
        trigger is reached.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            ------------------------------------------------
                              RHF COUPLED CLUSTER ITERATIONS
            ------------------------------------------------
        
        .. code-block:: text

            ------------------------------------------------
                              UHF COUPLED CLUSTER ITERATIONS
            ------------------------------------------------

        .. code-block:: text
        
            ----------------------------------------------------------
                                 OPEN-SHELL COUPLED CLUSTER ITERATIONS
            ----------------------------------------------------------
        """
        while 'ORCA POPULATION ANALYSIS' != line.strip() \
        and '*     ORCA property calculations      *' != line.strip():
            # Iter       E(tot)           E(Corr)          Delta-E          Residual     Time      <S|S>**1/2
            #  0     -7.469294707     -0.036553055      0.000000000      0.027013328    0.05      0.000000001
            #
            # or
            #
            # Iter       E(tot)           E(Corr)          Delta-E          Residual     Time
            #   0     -2.897443580     -0.035772194      0.000000000      0.027217829    0.00
            if line.strip()[:79] == 'Iter       E(tot)           E(Corr)          Delta-E          Residual     Time':
                # Extracts MP2 energies under the initial line.
                line = next(f)
                if 'energy_correl_mp2' not in self.parsed_info['outputs'].keys():
                    self.parsed_info['outputs']['energy_correl_mp2'] = []
                if 'mp2_total_energy' not in self.parsed_info['outputs'].keys():
                    self.parsed_info['outputs']['mp2_total_energy'] = []
                
                self.parsed_info['outputs']['energy_correl_mp2'].append(
                    float(line.split()[2])
                )
                self.parsed_info['outputs']['mp2_total_energy'].append(
                    float(line.split()[1])
                )
            
            # E(TOT)                                     ...     -7.473852176
            if line.strip()[:6] == 'E(TOT)':
                # Extracts total CCSD energy.
                if 'energy_ccsd' not in self.parsed_info['outputs'].keys():
                    self.parsed_info['outputs']['energy_ccsd'] = []
                
                self.parsed_info['outputs']['energy_ccsd'].append(
                    float(line.split()[2])
                )
            
            # T1 diagnostic                              ...      0.001316573 
            if line.strip()[:13] == 'T1 diagnostic':
                # Extracts T1 diagnostic.
                if 'diag_t1' not in self.parsed_info['outputs'].keys():
                    self.parsed_info['outputs']['diag_t1'] = []
                
                self.parsed_info['outputs']['diag_t1'].append(
                    float(line.split()[3])
                )
            
            # E(CCSD(T))                                 ...     -7.473882409
            if line.strip()[:10] == 'E(CCSD(T))':
                # Extracts total CCSD(T) energy..
                if 'energy_ccsd(t)' not in self.parsed_info['outputs'].keys():
                    self.parsed_info['outputs']['energy_ccsd(t)'] = []
                
                self.parsed_info['outputs']['energy_ccsd(t)'].append(
                    float(line.split()[2])
                )
            
            line = next(f)

    
    def scf_info(self, f, line):
        """Other scf information.

        This will be placed under the ``'keyword'`` JSON property.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.

        Returns
        -------
        :obj:`dict`
            Available SCF information that could contain the following keys.

            ``'approx_rij'``
                The resolution of identity (RI) approximation for the Coulomb
                (J) term.
            ``'approx_cosx'``
                The chain-of-spheres integration approximation to the exchange
                term (COSX).
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            ------------
            SCF SETTINGS
            ------------
        """
        while 'Total time needed     ' not in line.strip():
            # Hamiltonian:
            #  Ab initio Hamiltonian  Method          .... Hartree-Fock(GTOs)


            # General Settings:
            #  Integral files         IntName         .... al.chrg0.mult2-orca.sp.esp-ccsdt.anopvqz.vtightscf.sym-lambda0
            #  Hartree-Fock type      HFTyp           .... UHF
            if 'Ab initio Hamiltonian' == line.strip()[:21]:
                # We only include the HF type in the keywords.
                for _ in range(0, 5):
                    line = next(f)
                hf_type = line.split()[4]
                self.parsed_info['runtime_info']['hf_type'] = hf_type
            
            #  Number of Electrons    NEL             ....    3
            if 'Number of Electrons' == line.strip()[:19]:
                if 'n_ele' not in self.parsed_info['runtime_info'].keys():
                    self.parsed_info['runtime_info']['n_ele'] = []
                n_electrons = int(line.split()[5])
                self.parsed_info['runtime_info']['n_ele'].append(n_electrons)

            #  RI-approximation to the Coulomb term is turned on
            if 'RI-approximation to the Coulomb term is turned on' in line:
                self.parsed_info['runtime_info']['approx_rij'] = True

            #    RIJ-COSX (HFX calculated with COS-X)).... on
            if 'RIJ-COSX (HFX calculated with COS-X)' in line:
                self.parsed_info['runtime_info']['approx_cosx'] = True
            
            if 'RI-JK (J+K treated both via RI)' in line:
                self.parsed_info['runtime_info']['approx_rijk'] = True
            
            line = next(f)

    
    def mulliken_charges(self, f, line):
        """Mulliken atomic charges in same order as atomic coordinates.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            -----------------------
            MULLIKEN ATOMIC CHARGES
            -----------------------
        """
        line = self.skip_lines(f, 2)

        # Creates initial mulliken_charges property.
        if 'mulliken_charges' not in self.parsed_info['outputs'].keys():
            # self.parsed_info['outputs']['mulliken_charges'] = []
            pass
        
        # Appends Mulliken charges to a new item for every structure.
        # self.parsed_info['outputs']['mulliken_charges'].append([])
        while 'Sum of atomic charges' not in line:
            line_split = line.split(':')
            """
            self.parsed_info['outputs']['mulliken_charges'][-1].append(
                float(line_split[-1])
            )
            """
            line = next(f)

    
    def loewdin_charges(self, f, line):
        """Loewdin atomic charges in same order as atomic coordinates.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            ----------------------
            LOEWDIN ATOMIC CHARGES
            ----------------------
        """
        line = self.skip_lines(f, 2)

        # Creates initial loewdin_charges property.
        if 'loewdin_charges' not in self.parsed_info['outputs'].keys():
            # self.parsed_info['outputs']['loewdin_charges'] = []
            pass
        
        # Appends Loewdin charges to a new item for every structure.
        # self.parsed_info['outputs']['loewdin_charges'].append([])
        while '' != line.strip():
            line_split = line.split(':')
            """
            self.parsed_info['outputs']['loewdin_charges'][-1].append(
                float(line_split[-1])
            )
            """
            line = next(f)

    
    def dipole(self, f, line):
        """The X, Y, and Z dipole components.

        Final QCJSON specifies the method of the dipole moment (e.g.,
        ``'scf_dipole_moment'``, ``'mp2_dipole_moment'``). For now, we just
        store it as ``'dipole_moment'``.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            -------------
            DIPOLE MOMENT
            -------------
        """
        if 'dipole_moment' not in self.parsed_info['outputs'].keys():
            # self.parsed_info['outputs']['dipole_moment'] = []
            pass
        
        while 'Total Dipole Moment    :' not in line:
            line = next(f)
        line_split = line.split()
        dipole = [float(line_split[4]), float(line_split[5]), float(line_split[6])]
        #self.parsed_info['outputs']['dipole_moment'].append(dipole)

    
    def _add_geo_conv(self, info_label, line):
        """Parse and add geometric convergence info to data.

        Parameters
        ----------
        info_label : :obj:`str`
            Label for geometric convergence criteria.
        line : :obj:`str`
            Line from output file to extract information from.
        """
        split_line = line.split()
        value = float(split_line[2])
        target = float(split_line[3])
        if f'geo_{info_label}_target' not in self.parsed_info['runtime_info'].keys():
            self.parsed_info['runtime_info'][f'geo_{info_label}_target'] = target
        try:
            self.parsed_info['runtime_info'][f'geo_{info_label}_value'].append(value)
        except KeyError:
            self.parsed_info['runtime_info'][f'geo_{info_label}_value'] = [value]

    
    def geo_conv(self, f, line):
        """Extract geometric convergence values and tolerance.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

                                  .--------------------.
            ----------------------|Geometry convergence|-------------------------
            Item                value                   Tolerance       Converged
            ---------------------------------------------------------------------
        """
        while 'Max(Dihed)' not in line and 'Max(Improp)' not in line:
            if 'Energy change' in line:
                self._add_geo_conv('energy_change', line)
            elif 'RMS gradient' in line:
                self._add_geo_conv('rms_gradient', line)
            elif 'MAX gradient' in line:
                self._add_geo_conv('max_gradient', line)
            elif 'RMS step' in line:
                self._add_geo_conv('rms_step', line)
            elif 'MAX step' in line:
                self._add_geo_conv('max_step', line)
            line = next(f)

    
    def frequencies(self, f, line):
        """Extract vibrational frequencies, freq_vibs. Includes 0.00 frequencies.

        Based on https://github.com/MolSSI/QCSchema/pull/50#issuecomment-499155251.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            -----------------------
            VIBRATIONAL FREQUENCIES
            -----------------------
        """
        # Skips the following lines.
        # -----------------------
        # 
        # Scaling factor for frequencies =  1.000000000
        # 
        #    0:         0.00 cm**-1
        line = self.skip_lines(f, 5)

        # Sets up data.
        if 'freq_vib' not in self.parsed_info['outputs'].keys():
            self.parsed_info['outputs']['freq_vib'] = []

        vibfreqs = []
        while line.strip() != '':
            freq = float(line.split()[1])
            vibfreqs.append(freq)
            
            line = next(f)

        self.parsed_info['outputs']['freq_vib'].append(vibfreqs)

    
    def normal_modes(self, f, line):
        """Extract normalized, mass-weighted normal modes, q.

        Based on https://github.com/MolSSI/QCSchema/pull/50#issuecomment-499155251.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            ------------
            NORMAL MODES
            ------------
        """
        line = self.skip_lines(f, 8)

        # Sets up data.
        if 'q' not in self.parsed_info['outputs'].keys():
            self.parsed_info['outputs']['normal_modes'] = []

        q = []
        mode = 0
        while line.strip() != '-----------':
            if '.' not in line:
                # Skips lines that print column numbers. For example,
                #   0          1          2          3          4          5 
                # Also restarts mode index.
                mode = 0
            else:
                disp = [float(i) for i in line.split()[1:]]
                try:
                    q[mode].extend(disp)
                except IndexError:
                    q.append(disp)
                mode += 1
            line = next(f)
        self.parsed_info['outputs']['normal_modes'].append(q)

    
    def thermo_temp(self, f, line):
        """Extracts temperature used for thermochemistry.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            Temperature         ... 298.15 K
        """
        # Sets up data.
        if 'temp_thermochem' not in self.parsed_info['outputs'].keys():
            self.parsed_info['outputs']['temp_thermochem'] = []
        temp = float(line.split()[2])
        self.parsed_info['outputs']['temp_thermochem'].append(temp)
        # Needs to move off this line so extract can continue.
        line = next(f)

    
    def zpve(self, f, line):
        """Extract zero-point vibrational energy correction.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            Zero point energy                ...      0.06455884 Eh      40.51 kcal/mol
        """
        if 'correc_zpe' not in self.parsed_info['outputs'].keys():
            self.parsed_info['outputs']['correc_zpe'] = []
        zpve = float(line.split()[4])
        self.parsed_info['outputs']['correc_zpe'].append(zpve)
        # Needs to move off this line so extract can continue.
        line = next(f)

    
    def thermal_corrections(self, f, line):
        """Extracts thermal vibrational, rotational, and translational
        corrections.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            Total thermal correction                  0.00868494 Eh       5.45 kcal/mol
        """
        if 'correc_thermal' not in self.parsed_info['outputs'].keys():
            self.parsed_info['outputs']['correc_thermal'] = []
        thermal = float(line.split()[3])
        self.parsed_info['outputs']['correc_thermal'].append(thermal)
        line = next(f)  # Needs to move off this line so extract can continue.

    
    def enthalpic_corr(self, f, line):
        """Extract enthalpic corrections; the difference between H_298 and
        E_298.

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.

        .. code-block:: text

            Thermal Enthalpy correction       ...      0.00094421 Eh       0.59 kcal/mol
        """
        if 'correc_enthalpy' not in self.parsed_info['outputs'].keys():
            self.parsed_info['outputs']['correc_enthalpy'] = []
        enthalpy = float(line.split()[4])
        self.parsed_info['outputs']['correc_enthalpy'].append(enthalpy)
        line = next(f)  # Needs to move off this line so extract can continue.

    
    def entropic_corr(self, f, line):
        """Extract translational, rotational, and vibrational entropic
        corrections to enthalpy for Gibbs free energy (i.e., T*S_298).

        Parameters
        ----------
        f : :obj:`io.TextIOWrapper`
            Buffered text stream of the file.
        line : :obj:`str`
            Parsed line from ``f``.
        
        Notes
        -----
        Example trigger text for this extractor.
        
        .. code-block:: text

            Final entropy term                ...      0.04136711
        """
        if 'correc_entropy' not in self.parsed_info['outputs'].keys():
            self.parsed_info['outputs']['correc_entropy'] = []
        entropy = float(line.split()[4])
        self.parsed_info['outputs']['correc_entropy'].append(entropy)
        # Needs to move off this line so extract can continue.
        line = next(f)




# MIT License
# 
# Copyright (c) 2022, Alex M. Maldonado
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import hashlib
import numpy as np
import collections.abc

element_to_z = {
    'H': 1, 'He': 2, 'Li': 3, 'Be': 4, 'B': 5, 'C': 6, 'N': 7, 'O': 8, 'F': 9,
    'Ne': 10, 'Na': 11, 'Mg': 12, 'Al': 13, 'Si': 14, 'P': 15, 'S': 16,
    'Cl': 17, 'Ar': 18, 'K': 19, 'Ca': 20, 'Sc': 21, 'Ti': 22, 'V': 23,
    'Cr': 24, 'Mn': 25, 'Fe': 26, 'Co': 27, 'Ni': 28, 'Cu': 29, 'Zn': 30,
    'Ga': 31,'Ge': 32, 'As': 33, 'Se': 34, 'Br': 35, 'Kr': 36, 'Rb': 37,
    'Sr': 38, 'Y': 39, 'Zr': 40, 'Nb': 41, 'Mo': 42, 'Tc': 43, 'Ru': 44,
    'Rh': 45, 'Pd': 46, 'Ag': 47, 'Cd': 48, 'In': 49, 'Sn': 50, 'Sb': 51,
    'Te': 52, 'I': 53, 'Xe': 54, 'Cs': 55, 'Ba': 56, 'La': 57, 'Ce': 58,
    'Pr': 59, 'Nd': 60, 'Pm': 61, 'Sm': 62, 'Eu': 63, 'Gd': 64, 'Tb': 65,
    'Dy': 66, 'Ho': 67, 'Er': 68, 'Tm': 69, 'Yb': 70, 'Lu': 71, 'Hf': 72,
    'Ta': 73, 'W': 74, 'Re': 75, 'Os': 76, 'Ir': 77, 'Pt': 78, 'Au': 79,
    'Hg': 80, 'Tl': 81, 'Pb': 82, 'Bi': 83, 'Po': 84, 'At': 85, 'Rn': 86,
    'Fr': 87, 'Ra': 88, 'Ac': 89, 'Th': 90, 'Pa': 91, 'U': 92, 'Np': 93,
    'Pu': 94, 'Am': 95, 'Cm': 96, 'Bk': 97, 'Cf': 98, 'Es': 99, 'Fm': 100,
    'Md': 101, 'No': 102, 'Lr': 103, 'Rf': 104, 'Db': 105, 'Sg': 106,
    'Bh': 107, 'Hs': 108, 'Mt': 109, 'Ds': 110, 'Rg': 111, 'Cn': 112,
    'Uuq': 114, 'Uuh': 116,
}
z_to_element = {v: k for k, v in element_to_z.items()}

# Standard atomic weight from version 4.1 of
# https://www.nist.gov/pml/atomic-weights-and-isotopic-compositions-relative-atomic-masses
# If lower and upper bounds were provided the lower bound was selected.
z_to_mass = (
    None, 1.00784, 4.002602, 6.938, 9.0121831, 10.806, 12.0096, 14.00643,  # N
    15.99903, 18.998403163, 20.1797, 22.98976928, 24.304, 26.9815385, 28.084,  # Si
    30.973761998, 32.059, 35.446, 39.948, 39.0983, 40.078, 44.955908,  # Sc
    47.867, 50.9415, 51.9961, 54.938044, 55.845, 58.933194, 58.6934, 63.546,  # Cu
    65.38, 69.723, 72.630, 74.921595, 78.971, 79.901, 83.798, 85.4678, 87.62,  # Sr
    88.90584, 91.224, 92.90637, 95.95, 98.0, 101.07, 102.90550, 106.42, 107.8682,  # Ag
    112.414, 114.818, 118.710, 121.760, 127.60, 126.90447, 131.293,  # Xe
    132.90545196, 137.327, 138.90547, 140.116, 140.90766, 144.242, 145.0, 150.36,  # Sm
    151.964, 157.25, 158.92535, 162.500, 164.93033, 167.259, 168.93422, 173.054,  # Yb
    174.9668, 178.49, 180.94788, 183.84, 186.207, 190.23, 192.217, 195.084,  # Pt
    196.966569, 200.592, 204.382, 207.2, 208.98040, 209.0, 210.0, 222.0, 223.0, 226.0,  # Ra
    227.0, 232.0377, 231.03588, 238.02891, 237.0, 244.0  # Pu
)

def get_files(path, expression, recursive=True):
    """Returns paths to all files in a given directory that matches a provided
    expression in the file name.
    
    Commonly used to find all files of a certain type, e.g., output or xyz
    files.
    
    Parameters
    ----------
    path : :obj:`str`
        Specifies the directory to search.
    expression : :obj:`str`
        Expression to be tested against all file names in ``path``.
    recursive : :obj:`bool`, optional
        Recursively find all files in all subdirectories.
    
    Returns
    -------
    :obj:`list` [:obj:`str`]
        All absolute paths to files matching the provided expression.
    """
    if path[-1] != '/':
        path += '/'
    if recursive:
        all_files = []
        for (dirpath, _, filenames) in os.walk(path):
            index = 0
            while index < len(filenames):
                if dirpath[-1] != '/':
                    dirpath += '/'
                filenames[index] = dirpath + filenames[index]
                index += 1
            all_files.extend(filenames)
        files = []
        for f in all_files:
            if expression in f:
                files.append(f)
    else:
        files = []
        for f in os.listdir(path):
            filename = os.path.basename(f)
            if expression in filename:
                files.append(path + f)
    return files

def atoms_by_element(atom_list):
    """Converts a list of atoms identified by their atomic number to their
    elemental symbol in the same order.
    
    Parameters
    ----------
    atom_list : :obj:`list` [:obj:`int`]
        Atomic numbers of atoms within a structure.
    
    Returns
    -------
    :obj:`list` [:obj:`str`]
        Element symbols of atoms within a structure.
    """

    atom_list_elements = []
    for atom in atom_list:
        atom_list_elements.append(str(z_to_element[atom]))

    return atom_list_elements

def atoms_by_number(atom_list):
    """Converts a list of atoms identified by their elemental symbol to their
    atomic number.
    
    Parameters
    ----------
    atom_list : :obj:`list` [:obj:`str`]
        Element symbols of atoms within a structure.
    
    Returns
    -------
    :obj:`list` [:obj:`int`]
        Atomic numbers of atoms within a structure.
    """
    return [int(element_to_z[i]) for i in atom_list]

def parse_stringfile(stringfile_path):
    """Parses data from string file.

    A string file is data presented as consecutive xyz data. The data could be
    three Cartesian coordinates for each atom, three atomic force vector
    components, or both coordinates and atomic forces in one line (referred to
    as extended xyz).
    
    Parameters
    ----------
    stringfile_path : :obj:`str`
        Path to string file.
    
    Returns
    -------
    :obj:`tuple` [:obj:`list`]
        Parsed atoms (as element symbols :obj:`str`), comments, and data as
        :obj:`float` from string file.
    """
    Z, comments, data = [], [], []
    with open(stringfile_path, 'r') as f:
        for _, line in enumerate(f):
            line = line.strip()
            if not line:
                # Skips blank lines
                pass
            else:
                line_split = line.split()
                if len(line_split) == 1 \
                    and float(line_split[0]) % int(line_split[0]) == 0.0:
                    # Skips number of atoms line, adds comment line, and
                    # prepares next z and data item.
                    comment_line = next(f)
                    comments.append(comment_line.strip())
                    Z.append([])
                    data.append([])
                else:
                    # Grabs z and data information.
                    Z[-1].append(line_split[0])
                    data[-1].append([float(i) for i in line_split[1:]])
    return Z, comments, data

def get_md5(data, group_key, only_arrays=False, only_structures=False):
    """Creates MD5 hash for a group.

    Parameters
    ----------
    group : ``exdir.Object``
        An exdir object (``File`` or ``Group``)
    only_arrays : :obj:`bool`, optional
        Generate the MD5 hash using only arrays. This creates a data-centered
        MD5 that is not affected by data that are commonly added or
        changed. Defaults to ``False``.
    only_structures : :obj:`bool`, optional
        Generate the MD5 has with only ``atomic_numbers`` and ``geometry``
        if possible). This is more static than ``only_arrays`` and should be
        used to track sampling (i.e., ``r_prov_ids``).
    
    Returns
    -------
    :obj:`str`
        MD5 hash of a group.
    """
    md5_hash = hashlib.md5()
    # TODO: Figure out why different formats have different MD5s.

    if only_structures:
        try:
            Z = data.get(f'{group_key}/atomic_numbers')
            Z = Z.ravel()
            md5_hash.update(hashlib.md5(Z).digest())
        except Exception:
            pass
        try:
            R = data.get(f'{group_key}/geometry')
            R = R.ravel()
            md5_hash.update(hashlib.md5(R).digest())
        except Exception:
            pass
    else:
        keys = data.get_keys(group_key)
        for key in keys:
            if 'md5' in key:
                continue
            d = data.get(f'{group_key}/{key}')
            if isinstance(d, np.ndarray):
                d = d.ravel()
                md5_hash.update(hashlib.md5(d).digest())
            else:
                if only_arrays:
                    continue
                else:
                    md5_hash.update(repr(d).encode())
    
    return md5_hash.hexdigest()

def gen_entity_ids(
    atoms_per_mol, num_mol, starting_idx=0, add_to=None
):
    """Generates entity ids for a single species.

    Note that all of the atoms in each molecule must occur in the same order and
    be grouped together.

    Parameters
    ----------
    atoms_per_mol : :obj:`int`
        Number of atoms in the molecule.
    num_mol : :obj:`int`
        Number of molecules of this type in the system.
    starting_idx : :obj:`int`
        Number to start entity_id labels.
    add_to : :obj:`list`
        Entity ids to append new ids to.
    
    Returns
    -------
    :obj:`numpy.ndarray`
        Entity ids for a structure.
    """
    entity_ids = []
    for i in range(starting_idx, num_mol+starting_idx):
        entity_ids.extend([i for _ in range(0, atoms_per_mol)])
    
    if add_to is not None:
        if isinstance(add_to, np.ndarray):
            add_to = add_to.tolist()
        return np.array(add_to + entity_ids)
    else:
        return np.array(entity_ids)

def gen_comp_ids(label, num_mol, entity_ids, add_to=None):
    """Prepares the list of component ids for a system with only one species.

    Parameters
    ----------
    label : :obj:`int`
        Species label.
    num_mol : :obj:`int`
        Number of molecules of this type in the system.
    entity_ids : :obj:`int`
        Entity ids for these species.
    add_to : :obj:`list`
        Component ids to append new ids to.

    Returns
    -------
    :obj:`numpy.ndarray`
        Component ids for a structure.
    """
    comp_ids = [label for _ in range(0, num_mol)]
    if add_to is not None:
        if isinstance(add_to, np.ndarray):
            add_to = add_to.tolist()
        return np.array(add_to + comp_ids)
    else:
        return np.array(comp_ids)

def center_structures(Z, R):
    """Centers each structure's center of mass to the origin.

    Previously centered structures should not be affected by this technique.

    Parameters
    ----------
    Z : :obj:`numpy.ndarray`
        Atomic numbers of the atoms in every structure.
    R : :obj:`numpy.ndarray`
        Cartesian atomic coordinates of data set structures.
    
    Returns
    -------
    :obj:`numpy.ndarray`
        Centered Cartesian atomic coordinates.
    """
    # Masses of each atom in the same shape of R.
    if R.ndim == 2:
        R = np.array([R])
    
    masses = np.empty(R[0].shape)
    
    for i in range(len(masses)):
        masses[i,:] = z_to_mass[Z[i]]
    
    for i in range(len(R)):
        r = R[i]
        cm_r = np.average(r, axis=0, weights=masses)
        R[i] = r - cm_r
    
    if R.shape[0] == 1:
        return R[0]
    else:
        return R

def combine_dicts(dict1, dict2):
    """Combine two dictionaries.
    
    Parameters
    ----------
    dict1 : :obj:`dict`
    dict2 : :obj:`dict`

    Returns
    -------
    :obj:`dict`
    """
    for k, v in dict2.items():
        if isinstance(v, collections.abc.Mapping):
            dict1[k] = combine_dicts(dict1.get(k, {}), v)
        else:
            dict1[k] = v
    return dict1



# MIT License
# 
# Copyright (c) 2022, Alex M. Maldonado
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
import psi4

def psi4_engrad(Z, R, R_idxs, charge, mult, method, options, threads):
    """Ray remote function for computing total electronic energy and atomic
    gradients using Psi4.

    Parameters
    ----------
    Z : ``ray.ObjectRef`` of :obj:`numpy.ndarray`
        Atomic numbers of the atoms with repsect to ``R``.
    R : ``ray.ObjectRef`` of :obj:`numpy.ndarray`
        Cartesian coordinates of all structures in group. This includes
        unused structures.
    R_idxs : :obj:`numpy.ndarray`
        Indices of the structures from ``R`` to compute energies and gradients
        for.
    charge : :obj:`int`
        Total molecular charge.
    mult : :obj:`int`
        Total molecular multiplicity.
    method : :obj:`str`
        Specifies the Psi4 method used for the gradient. For more information,
        please see `the Psi4 documentation <https://psicode.org/psi4manual/\
        master/opt.html#geometry-optimization-w-w-optimize-and-gradient>`_ for
        your specific version.
    options : :obj:`dict`
        `Psi4 control keywords <https://psicode.org/psi4manual/master\
        /psithoninput.html#job-control-keywords>`_ using the PsiAPI format.
        Some common ones are ``basis``, ``e_convergence``, ``d_convergence``,
        and ``reference``.
    threads : :obj:`int`
        Number of threads for Psi4. This is almost always the number of cores
        being used for the worker.
    
    Returns
    -------
    :obj:`numpy.ndarray`
        ``R_idxs``
    :obj:`numpy.ndarray`
        Total electronic energy of computed structures in the same order as
        ``R_idxs``. Units of Hartree.
    :obj:`numpy.ndarray`
        Atomic gradients of computed structures in the same order as ``R_idxs``.
        Units of Hartree/Angstrom.
    """
    psi4.core.set_num_threads(threads)
    psi4.set_options(options)
    R = R[R_idxs]
    G = np.zeros(R.shape)
    E = np.zeros(R.shape[0])
    for i in range(len(R)):
        mol = psi4.core.Molecule.from_arrays(
            geom=R[i], elez=Z, molecular_charge=charge,
            molecular_multiplicity=mult
        )
        g, wfn_mp2 = psi4.gradient(
            method, molecule=mol, return_wfn=True
        )
        g = g.to_array()
        g /= psi4.constants.bohr2angstroms
        G[i] = g
        E[i] = wfn_mp2.energy()
    return R_idxs, E, G



# MIT License
# 
# Copyright (c) 2022, Alex M. Maldonado
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from random import randrange, sample, choice
import itertools
import numpy as np
from .utils import center_structures as get_center_structures
from .utils import get_md5
from . import _version
__version__ = _version.get_versions()['version']

def _get_R_prov_id(
    dest_R_prov_ids, source_md5
):
    """Determines the structure_prov_id for the sampling source.

    It is possible that 

    Parameters
    ----------
    dest_R_prov_ids : :obj:`dict`
        The r_prov_ids
    source_md5 : :obj:`str`
        
    
    Returns
    -------
    :obj:`int`
        Numerical ID for the Rset.
    """
    R_prov_id = None
    if dest_R_prov_ids == {}:
        # No previous sampling, so this automatically becomes zero.
        R_prov_id = 0
    else:
        # Check if md5 is already stored.
        for key,value in dest_R_prov_ids.items():
            if value == source_md5:
                R_prov_id = key
                break
        # Generate a new R_prov_id because it was not already stored.
        if R_prov_id is None:
            R_prov_id = max(dest_R_prov_ids.keys()) + 1

    return R_prov_id

def _initialize_structure_sampling_arrays(
    Z, R, E, G, R_prov_specs, quantity, comp_labels, R_source, entity_ids_source,
    entity_ids_samples, copy_EG
):
    """Creates or extends arrays for sampling structures.

    Parameters
    ----------
        
    
    Returns
    -------

    """
    # Either creates new ones or concatenates to
    # original destination data.
    # Trial selection to get shapes for R and G
    if quantity == 'all':
        # Determine the maximum amount of structures we can sample from a single
        # source structure.
        combs_per_structure = sum(
            1 for _ in itertools.product(*entity_ids_samples)
        )
        size_quantity = int(combs_per_structure*len(R_source))
        sampling_all = True
    else:
        size_quantity = int(quantity)
        sampling_all = False
    
    # Determine number of atoms.
    entity_id_for_size = (i[0] for i in entity_ids_samples)
    number_of_atoms = 0
    for entity_id in entity_id_for_size:
        number_of_atoms += len(np.where(entity_ids_source==entity_id)[0])
    
    if Z is None:
        Z = np.zeros((number_of_atoms,))
    else:
        assert len(Z) == number_of_atoms
    
    sampled_R_empty = np.empty((size_quantity, number_of_atoms, 3))
    sampled_R_empty[:] = np.nan
    if R is None:
        idx_selection = 0  # Starting index of the newley sampled structures.
        R = sampled_R_empty
    else:
        idx_selection = len(R)
        R = np.concatenate((R, sampled_R_empty), axis=0)
    
    if copy_EG:
        sampled_E_empty = np.empty(sampled_R_empty.shape[0])
        sampled_E_empty[:] = np.nan
        if E is None:
            E = sampled_E_empty
        else:
            E = np.concatenate((E, sampled_E_empty))
        
        sampled_G_empty = np.empty((size_quantity, number_of_atoms, 3))
        sampled_G_empty[:] = np.nan
        if G is None:
            G = sampled_G_empty
        else:
            G = np.concatenate((G, sampled_G_empty), axis=0)
    else:
        E = None
        G = None
    
    sampled_R_prov_specs_empty = np.empty((size_quantity, 2+len(comp_labels)))
    sampled_R_prov_specs_empty[:] = np.nan
    if R_prov_specs is None:
        R_prov_specs = sampled_R_prov_specs_empty
    else:
        R_prov_specs = np.concatenate(
            (R_prov_specs, sampled_R_prov_specs_empty), axis=0
        )
    
    return (Z, R, E, G, R_prov_specs, sampling_all, idx_selection)

def _generate_structure_samples(
    quantity, structure_idxs, entity_ids_samples
):
    """Randomly generates structure and entity selections.

    The prov ID for the structure is not included here and must be
    inserted later.

    Parameters
    ----------
    quantity : :obj:`int` or :obj:`float`
        Number of structures to sample from the structure set. For example,
        ``'100'``, ``452``, or even ``'all'``.
    structure_idxs : :obj:`numpy.ndarray`
        Indices of source structures we can sample from.
    entity_ids_samples : :obj:`tuple` (:obj:`tuple`)
        A nested tuple containing all possible entity selections for all desired
        components of the sampled structure. Has length ``n``, where ``n`` is
        the number of entities in the sampled structure. Each element is another
        :obj:`tuple` containing all possible source ``entity_ids`` for that
        destination ``entity_id``.

    Yields
    ------
    :obj:`list`
        Structure index and ``entity_id`` selections.

    Notes
    -----
    This could generate selections of multiple of the same entities.
    For example, ``[602, 0, 70, 70]`` selects entity ``70`` twice which is
    nonphysical. For now we will offload this check to 
    """
    # Sampling a specified number of structures.
    if isinstance(quantity, int) or str(quantity).isdigit():
        while True:
            R_selection_idx = randrange(len(structure_idxs))
            entity_selection = [None for entity_ids in entity_ids_samples]
            for i in range(len(entity_selection)):
                entity_selection[i] = choice(
                    tuple(
                        filter(
                            lambda x: x not in entity_selection[:i+1],
                            entity_ids_samples[i]
                        )
                    )
                )
            selection = [R_selection_idx] + entity_selection
            yield selection
    # Sampling all possible structures.
    elif quantity == 'all':
        for R_selection_idx in structure_idxs:
            #all_combinations = unique_prod(*entity_ids_samples)
            for comb in itertools.product(*entity_ids_samples):
                if len(comb) != len(set(comb)):
                    continue
                selection = [R_selection_idx] + list(comb)
                yield selection

def sample_structures(
    source_data, source_key, quantity, comp_labels, R_prov_ids, source_R_prov_specs,
    Z=None, R=None, R_prov_specs=None, structure_idxs=None,
    criteria=None, z_slice=None, cutoff=None, center_structures=False,
    sampling_updates=False, copy_EG=False, E=None, G=None,
    energy_label_source='energy_ele', grad_label_source='grads'
):
    """Randomly samples structures from a source.

    If previous data are already provided (i.e., ``Z``, ``R``, and
    ``R_prov_specs`` are not ``None``) then the sampled structures are appended
    to these arrays

    Parameters
    ----------
    source_data : ``exdir.Group``
        A loaded ``exdir.Group`` to sample structures from. Must contain
        ``z``, ``r``, ``entity_ids``, and
        ``comp_ids``.
    source_key : :obj:`str`
        Key to the desired source group.
    quantity : :obj:`str`
        Number of structures to sample from the structure set. For example,
        ``'100'``, ``'452'``, or even ``'all'``.
    comp_labels : :obj:`tuple`
        The ``component_id`` labels to include in sample structures. The order
        of each label does matter.
    R_prov_ids : :obj:`dict` {:obj:`int`: :obj:`str`}
        Structure prov IDs including any originally defined IDs and new
        ones being added from the source.
    source_R_prov_specs : :obj:`numpy.ndarray` or ``None``
        Structure prov specifications of the source if it was created by
        sampling from other sources. If the ``source`` is original, this should
        be ``None``.
    Z : :obj:`numpy.ndarray`, optional
        Atomic numbers of destination if they already exist. Defaults 
        to ``None``.
    R : :obj:`numpy.ndarray`, optional
        Cartesian atomic coordinates of destination if they already exist.
        Defaults to ``None``.
    R_prov_specs : :obj:`numpy.ndarray`, optional
        Structure prov specifications of destination. Defaults to
        ``None``.
    structure_idxs : :obj:`tuple`, optional
        Source indices of structures to sample from. Defaults to sampling from
        all structures.
    criteria : ``func``, optional
        Structure criteria during the sampling procedure. Defaults to
        ``None`` if no criteria should be used.
    z_slice : :obj:`numpy.ndarray`, optional
        Indices of the atoms to be used for the cutoff calculation. Defaults
        to ``[]`` is no criteria is selected or if it is not required for
        the selected criteria.
    cutoff : :obj:`list`, optional
        Distance cutoff between the atoms selected by ``z_slice``. Must be
        in the same units (e.g., Angstrom) as ``R``. Defaults to ``None`` if
        no criteria is selected or a cutoff is not desired.
    center_structures : :obj:`bool`, optional
        Move the center of mass of each structure to the origin thereby 
        centering each structure (and losing the actual coordinates of
        the structure). Defaults to ``False``.
    sampling_updates : :obj:`bool`, optional
        Will print something for every 100 successfully sampled structures.
        Defaults to ``False``.
    copy_EG : :obj:`bool`, optional
        Creates datasets for energies and gradients (using ``energy_label`` and 
        ``grad_label``) and attempts to copy data from the source if possible.
        If no compatible data is available it will just store :obj:`numpy.NaN`.
        Defaults to ``True``.
    E : :obj:`numpy.ndarray`, optional
        Original energies, if any, to append to. Defaults to ``None``.
    G : :obj:`numpy.ndarray`, optional
        Original atomic gradients, if any, to append to. Defaults to ``None``.
    energy_label_source : :obj:`str`, optional
        Specifies source dataset name containing energies to copy.
        Defaults to ``'energy_ele'``. If ``copy_EG`` is ``True``, but
        there are no valid energy datasets then we create an empty array.
    grad_label_source : :obj:`str`, optional
        Specifies source dataset name containing atomic gradients to copy.
        Defaults to ``'grads'``. If ``copy_EG`` is ``True``, but
        there are no valid energy datasets then we create an empty array.
    
    Returns
    -------
    :obj:`numpy.ndarray`
        Atomic coordinates of sampled structures.
    :obj:`numpy.ndarray`
        Atomic coordinates of sampled structures appended to any original
        structures.
    :obj:`numpy.ndarray` or ``None``
        Energies of sampled structures appended to any original structures.
        ``None`` if ``copy_EG`` is ``False``.
    :obj:`numpy.ndarray` or ``None``
        Atomic gradients of sampled structures appended to any original structures.
        ``None`` if ``copy_EG`` is ``False``.
    :obj:`numpy.ndarray`
        ``entity_ids`` of the structures.
    :obj:`numpy.ndarray`
        ``R_prov_specs`` of the structures.
    """
    # Get data from structure source.
    Z_source = source_data.get(f'{source_key}/atomic_numbers')
    R_source = source_data.get(f'{source_key}/geometry')
    entity_ids_source = source_data.get(f'{source_key}/entity_ids')
    comp_ids_source = source_data.get(f'{source_key}/comp_ids')

    # If source_R_prov_specs is None, then this is an original source.
    # Which means there should only be one R_prov_id (we store this for later).
    if source_R_prov_specs is None:
        assert len(R_prov_ids) == 1
        source_R_prov_id = tuple(R_prov_ids.keys())[0]

    if copy_EG:
        try:
            E_source = source_data.get(f'{source_key}/{energy_label_source}')
            G_source = source_data.get(f'{source_key}/{grad_label_source}')
        except Exception:
            E_source = None
            G_source = None
    else:
        E_source = None
        G_source = None


    ### Setup sampling   ###

    # TODO: Setup option to have one entity be the same all the time.

    ###   Sampling setup   ###
    # Determine what entity IDs we can sample from for each component.
    comp_ids_source_array = np.array(comp_ids_source)
    entity_ids_samples = tuple(
        (
            tuple(np.argwhere(comp_ids_source_array == comp_id)[:,0]) \
            for comp_id in comp_labels
        )
    )
    
    # Check that all atomic numbers of each entity is consistent.
    # Check that the z for these new samples match the destination.
    # Also build the entity_ids of these structures.
    Z_sample = []
    entity_ids = []
    entity_id = 0
    for dest_entity in entity_ids_samples:
        entity_ref = dest_entity[0]
        Z_ref = Z_source[np.argwhere(entity_ids_source == entity_ref)[:,0]]
        entity_ids.extend([entity_id for _ in range(len(Z_ref))])
        entity_id += 1
        Z_sample.extend(Z_ref)
        if len(dest_entity) != 1:
            for other_entity in dest_entity[1:]:
                Z_other = Z_source[np.argwhere(entity_ids_source == other_entity)[:,0]]
                try:
                    assert np.array_equal(Z_ref, Z_other)
                except AssertionError:
                    e = f'Atomic numbers do not match for entity_id of {other_entity}'
                    raise AssertionError(e)
    entity_ids = np.array(entity_ids)
    
    Z_sample = np.array(Z_sample)
    if Z is not None:
        try:
            assert np.array_equal(Z, Z_sample)
        except AssertionError:
            print(f'Destination atomic numbers: {Z}')
            print(f'Sample atomic numbers: {Z_sample}')
            e = 'Atomic numbers of samples do not match destination.'
            raise AssertionError(e)
    else:
        Z = Z_sample


    ###   Initialize all arrays   ###
    Z, R, E, G, R_prov_specs, sampling_all, idx_selection = \
        _initialize_structure_sampling_arrays(
        Z, R, E, G, R_prov_specs, quantity, comp_labels, R_source,
        entity_ids_source, entity_ids_samples, copy_EG
    )

    ###   Sampling counters   ###
    # TODO: clean up progress printing (look into MDAnalysis rdf)
    num_generated = 0  # Number of structures we generated (includes rejected).
    num_accepted = 0  # Number of structures successfully sampled.
    prev_num_accepted_print = 0  # Workaround to control printing.
    
    ###   Begin sampling structures   ###
    if structure_idxs is None:
        structure_idxs = tuple(range(0, len(R_source)))
    for selection in _generate_structure_samples(
        quantity, structure_idxs, entity_ids_samples
    ):
        num_generated += 1

        ###   Sampling maintenance   ###
        # Ends sampling for number quantities.
        if not sampling_all:
            if num_accepted == int(quantity):
                break

        ###   Sampling updates   ###
        # Prints progress information every 500 successful samples.
        if sampling_updates:
            if not sampling_all:
                if num_accepted%500 == 0 \
                    and num_accepted != prev_num_accepted_print:
                    prev_num_accepted_print = num_accepted
                    print(f'Successfully sampled {num_accepted} structures')
            else:
                if (num_accepted+1)%500 == 0:
                    print(
                        f'Successfully sampled {num_accepted+1} structures'
                    )
        
        ###   Handle r_prov_specs   ###
        # Determines the correct prov specification based on the sample and source.
        # We have already determined source_R_prov_id for an original source.
        # Just need to retrieve the source_R_prov_id of a source that is
        # from sampled structures.
        if source_R_prov_specs is not None:
            source_R_prov_id = source_R_prov_specs[selection[0]][0]
        selection.insert(0, source_R_prov_id)


        # Check if selection is already in the destination.
        # If it is, we do not include this selection and try again.
        if (R_prov_specs[:idx_selection]==selection).all(1).any():
            continue
        
        ###   Checks structure criteria   ###
        # Creates mask for atoms in the selection.
        r_index_source = selection[1]
        atom_idx_mask_gen = (
            entity_ids_source == entity_id for entity_id in selection[2:]
        )
        atom_idx_mask = np.bitwise_or.reduce(
            np.array(list(atom_idx_mask_gen)), axis=0
        )

        entity_ids_selection = entity_ids_source[atom_idx_mask]
        r_selection = R_source[r_index_source][atom_idx_mask]
        if criteria is not None:
            # TODO: Update criteria to test for multiple structure (i.e., 3 dimensions)
            accept_r, _ = criteria(
                Z, r_selection, z_slice, entity_ids_selection, cutoff
            )
            # If criteria is not met, will not include sample.
            if not accept_r:
                continue
        
        
        ###   SUCCESSFUL SAMPLE   ###
        R[idx_selection] = r_selection
        R_prov_specs[idx_selection] = selection
        if copy_EG:
            if E_source is not None:
                E[idx_selection] = E_source[r_index_source]
            if G_source is not None:
                G[idx_selection] = G_source[r_index_source][atom_idx_mask]
        
        num_accepted += 1
        idx_selection += 1
    
    if copy_EG:
        E = E[:idx_selection]
        G = G[:idx_selection]

    # Center structures by moving the center of mass to the origin.
    if center_structures:
        R = get_center_structures(Z, R)
    
    return (
        Z, R[:idx_selection], E, G, entity_ids, R_prov_specs[:idx_selection]
    )

def add_structures_to_group(
    source_data, source_key, dest_data, dest_key, quantity,
    comp_labels, structure_idxs=None, consistent_entities=None, criteria=None,
    z_slice=[], cutoff=[], center_structures=False, sampling_updates=False,
    copy_EG=False, energy_labels=('energy_ele',), grad_labels=('grads',),
    write=True
):
    """Adds randomly sampled structures to a group.

    Parameters
    ----------
    source_data : ``reptar.data``
        A reptar data object to sample structures from. Must contain
        ``atomic_numbers``, ``geometry``, ``entity_ids``, and ``comp_ids``.
        If it does not contain a ``md5_structures`` one will be created and
        saved.
    source_key : :obj:`str`
        Key to the desired source group.
    dest_data : ``reptar.data``
        A reptar data object to add sampled structures to.
    dest_key : :obj:`str`
        Key to the desired destination group. If it does not
        exist then it will be created.
    quantity : :obj:`str` or :obj:`int`
        Number of structures to sample from the data. For example,
        ``'100'``, ``452``, or ``'all'``.
    comp_labels : :obj:`tuple`
        The component ID labels to include in sample structures. The order of
        each label does matter.
    structure_idxs : :obj:`tuple`, optional
        Specific structure indices from source to sample from. Defaults to all
        structures.
    consistent_entities : :obj:`list` [:obj:`int`], optional
        Molecule indices that will be in every selection. Not implemented
        yet. Defaults to ``None``.
    criteria : ``object``, optional
        Structure criteria during the sampling procedure. Defaults to
        ``None`` if no criteria should be used.
    z_slice : :obj:`numpy.ndarray`, optional
        Indices of the atoms to be used for the cutoff calculation. Defaults
        to ``[]`` is no criteria is selected or if it is not required for
        the selected criteria.
    cutoff : :obj:`list`, optional
        Distance cutoff between the atoms selected by ``z_slice``. Must be
        in the same units (e.g., Angstrom) as ``R``. Defaults to ``[]`` if
        no criteria is selected or a cutoff is not desired.
    center_structures : :obj:`bool`, optional
        Move the center of mass of each structure to the origin thereby 
        centering each structure (and losing the actual coordinates of
        the structure). Defaults to ``False``.
    sampling_updates : :obj:`bool`, optional
        Will print something for every 100 successfully sampled structures.
        Defaults to ``False``.
    copy_EG : :obj:`bool`, optional
        Creates datasets for energies and gradients (using ``energy_label`` and 
        ``grad_labels``) and attempts to copy data from the source if possible.
        If no compatible data is available it will just store :obj:`numpy.NaN`.
        Defaults to ``True``.
    energy_labels : :obj:`tuple` (:obj:`str`), optional
        Specifies all energy keys in the destination group to extend with the
        new sampling. The array from each key will have ``quantity`` amount of 
        ``np.NaN`` values appended to it. The first value should be the desired
        key from the source. Defaults to ``('energy_ele',)``.
    grad_labels : :obj:`str` (:obj:`str`), optional
        Specifies all gradient keys in the destination group to extend with the
        new sampling. The array 1from each key will have ``quantity`` amount of 
        ``np.NaN`` values appended to it. The first value should be the desired
        key from the source. Defaults to ``('grads',)``.
    write : :obj:`bool`, optional
        Write newly sampled to ``Group``. Defaults to True
    
    Notes
    -----
    Nothing is saved until all checks are passed to avoid irreversible changes
    to already existing ``Group``.
    """
    # Grabs data from destination if exists.
    try:
        Z = dest_data.get(f'{dest_key}/atomic_numbers')
    except Exception:
        Z = None
    
    try:
        R = dest_data.get(f'{dest_key}/geometry')
    except Exception:
        R = None
    
    # Get energies and gradients if available.
    # TODO: perform check if we should copy energies and gradients based on if the
    # sampled structure is the entire source structure.
    if copy_EG:
        E = dest_data.get(f'{dest_key}/{energy_labels[0]}')
        G = dest_data.get(f'{dest_key}/{grad_labels[0]}')
    else:
        E = None
        G = None

    try:
        R_prov_ids = dest_data.get(f'{dest_key}/r_prov_ids')
        R_prov_specs = dest_data.get(f'{dest_key}/r_prov_specs')
    except Exception as e:
        R_prov_ids = None
        R_prov_specs = None
    
    try:
        entity_ids = dest_data.get(f'{dest_key}/entity_ids')
        comp_ids = dest_data.get(f'{dest_key}/comp_ids')
    except Exception:
        entity_ids = None
        comp_ids = None
    
    # Handle if source already has r_prov_ids
    try:
        source_R_prov_ids = source_data.get(f'{source_key}/r_prov_ids')
        source_R_prov_specs = source_data.get(f'{source_key}/r_prov_specs')
    except Exception as e:
        # This source is an original (was not created from sampling).
        source_R_prov_ids = None
        source_R_prov_specs = None
    
    ###   Check component IDs   ###
    if comp_ids is not None:
        try:
            assert np.array_equal(comp_ids, np.array(comp_labels))
        except AssertionError:
            e = f'Component IDs of destination ({comp_ids.tolist()}) do not match comp_labels ({comp_labels}).'
            raise AssertionError(e)
    else:
        comp_ids = np.array(comp_labels)


    # Prepare source r_prov_ids and r_prov_specs.
    # If source is an original, we just need to determine the next prov_id
    # and prov_specs stays None.
    # If the source was created with sampling, we need to adjust the
    # prov_ids and prov_specs to not overlap with the destination.
    # For example, change 0 from source to 1 if 0 is already taken in destination.

    # Original source (not from sampled)
    if source_R_prov_ids is None:
        try:
            md5_source = source_data.get(f'{source_key}/md5_structures')
        except Exception:
            md5_source = get_md5(source_data, source_key, only_structures=True)
            source_data.add(f'{source_key}/md5_structures', md5_source)
        
        # Create pseudo source_R_prov_ids.
        source_R_prov_ids = {0: md5_source}

    # If is this not a new destination we cannot always reuse the source information.
    # Need to check for overlap of ids and specifications.
    # Just need to shift the new source IDs up by the maximum destination ID.
    if R_prov_ids is not None:

        # Check and remove any IDs already in the destination.
        present_ids = tuple(key for key in R_prov_ids.keys())
        present_md5s = tuple(value for value in R_prov_ids.values())
        remove_source_ids = []
        for source_id,source_md5 in source_R_prov_ids.items():
            if source_md5 in present_md5s:
                update_id = present_ids[present_md5s.index(source_md5)]
                if update_id != source_id:
                    if source_R_prov_specs is not None:
                        update_idx = np.argwhere(
                            source_R_prov_specs[:,0] == source_id
                        )[:,0]
                        source_R_prov_specs[update_idx,0] = update_id
                remove_source_ids.append(source_id)
        if len(remove_source_ids) > 0:
            for source_id in remove_source_ids:
                del source_R_prov_ids[source_id]
        
        # Create any new source IDs that will be added to destination.
        if len(source_R_prov_ids) > 0:
            max_dest_prov_id = max(R_prov_ids.keys())
            next_id = max_dest_prov_id + 1

            for source_id in tuple(source_R_prov_ids.keys()):
                source_R_prov_ids[next_id] = source_R_prov_ids.pop(
                    source_id
                )
                if source_R_prov_specs is not None:
                    update_idx = np.argwhere(
                        source_R_prov_specs[:,0] == source_id
                    )[:,0]
                    source_R_prov_specs[update_idx,0] = next_id
                next_id += 1
    
            # Merge the IDs
            new_R_prov_ids = {**R_prov_ids, **source_R_prov_ids}
        else:
            new_R_prov_ids = R_prov_ids
    else:
        new_R_prov_ids = source_R_prov_ids
    
    # Stores the index of the first newly sampled structure for checks.
    if R is None:
        n_R_initial = 0
    else:
        n_R_initial = R.shape[0]
    
    # Begin sampling.
    Z, R, E, G, entity_ids_sampled, R_prov_specs = sample_structures(
        source_data, source_key, quantity, comp_labels, new_R_prov_ids,
        source_R_prov_specs, Z=Z, R=R, R_prov_specs=R_prov_specs,
        structure_idxs=structure_idxs, criteria=criteria, z_slice=z_slice,
        cutoff=cutoff, center_structures=center_structures,
        sampling_updates=sampling_updates, copy_EG=copy_EG, E=E, G=G,
        energy_label_source=energy_labels[0], grad_label_source=grad_labels[0]
    )

    num_sampled = R.shape[0] - n_R_initial

    # Check that entity_ids are the same.
    if entity_ids is not None:
        try:
            assert np.array_equal(entity_ids, entity_ids_sampled)
        except AssertionError:
            e = f'Destination entity_ids ({entity_ids.tolist()}) do not '\
                f'match sampled entity_ids ({entity_ids_sampled.tolist()}).'
            raise AssertionError(e)
    else:
        entity_ids = entity_ids_sampled
    
    if write:
        dest_data.add(f'{dest_key}/atomic_numbers', Z)
        dest_data.add(f'{dest_key}/geometry', R)
        if copy_EG:
            dest_data.add(f'{dest_key}/{energy_labels[0]}', E)
            dest_data.add(f'{dest_key}/{grad_labels[0]}', G)
            if len(energy_labels) > 1:
                for i in range(1, len(energy_labels)):
                    e_label = energy_labels[i]
                    e_data = dest_data.get(f'{dest_key}/{e_label}')
                    e_data_shape = e_data.shape
                    e_shape_new = (e_data_shape[0]+num_sampled,)
                    e_data_new = np.empty(e_shape_new)
                    e_data_new[:] = np.nan
                    e_data_new[:e_data_shape[0]] = e_data
                    dest_data.add(f'{dest_key}/{e_label}', e_data_new)
                for i in range(1, len(grad_labels)):
                    g_label = grad_labels[i]
                    g_data = dest_data.get(f'{dest_key}/{g_label}')
                    g_data_shape = g_data.shape
                    g_shape_new = (
                        g_data_shape[0]+num_sampled, g_data_shape[1],
                        g_data_shape[2]
                    )
                    g_data_new = np.empty(g_shape_new)
                    g_data_new[:] = np.nan
                    g_data_new[:g_data_shape[0]] = g_data
                    dest_data.add(f'{dest_key}/{g_label}', g_data_new)
        dest_data.add(f'{dest_key}/entity_ids', entity_ids)
        dest_data.add(f'{dest_key}/r_prov_ids', new_R_prov_ids)
        dest_data.add(f'{dest_key}/r_prov_specs', R_prov_specs)
        dest_data.add(f'{dest_key}/r_centered', center_structures)
        dest_data.add(f'{dest_key}/comp_ids', comp_ids)

        dest_data.add(
            f'{dest_key}/md5',
            get_md5(dest_data, dest_key)
        )
        dest_data.add(
            f'{dest_key}/md5_structures', get_md5(
                dest_data, dest_key, only_arrays=True
            )
        )
        dest_data.add(f'{dest_key}/reptar_version', __version__)

    return dest_data



# MIT License
# 
# Copyright (c) 2022, Alex M. Maldonado
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import numpy as np
import json
import exdir
import cclib
from .utils import combine_dicts

class data:
    """Create, store, and access data from a variety of formats.

    Parameters
    ----------
    file_path : :obj:`str`
        Path to a file supported by reptar. If it does not exist, then one will
        be created if possible.
    mode : :obj:`str`, optional
        A file mode string that defines the read/write behavior. Defaults to
        ``'r'``.
    allow_remove : :obj:`bool`, optional
        Allow the removal of exdir groups in ``w`` operation. Defaults to
        ``False``.
    plugins : :obj:`list`, optional
        A list of instantiated exdir plugins. Defaults to ``None``.
    from_dict : :obj:`dict`, optional
        Load data from a dictionary.
    """

    def __init__(self, file_path, mode='r', allow_remove=False,
        plugins=None, from_dict=None
    ):
        if from_dict is None:
            self._from_path(
                file_path, mode, allow_remove, plugins
            )
        else:
            self._from_dict(
                file_path, from_dict, mode, allow_remove, plugins
            )
    
    def _from_path(
        self, file_path, mode, allow_remove, plugins
    ):
        """Populates the data object from a file path.

        Parameters
        ----------
        file_path : :obj:`str`
            Path to a file supported by reptar. If it does not exist, then one will
            be created if possible.
        mode : :obj:`str`, optional
            A file mode string that defines the read/write behavior. Defaults to
            ``'r'``.
        allow_remove : :obj:`bool`, optional
            Allow the removal of exdir groups in ``w`` operation. Defaults to
            ``False``.
        plugins : :obj:`list`, optional
            A list of instantiated exdir plugins. Defaults to ``None``.
        """
        exists = os.path.exists(file_path)
        _, f_ext = os.path.splitext(file_path)

        if f_ext == '.exdir':
            File = exdir.File(
                file_path, mode, allow_remove, plugins
            )
        elif f_ext == '.json':
            if exists:
                with open(file_path, 'r') as f:
                    File = json.load(f)
            else:
                File = {}
        elif f_ext == '.npz':
            if exists:
                File = dict(np.load(file_path, allow_pickle=True))
                # Since everything is stored in arrays, we clean up array data.
                for k,v in File.items():
                    if self._is_iter(v):
                        v = self.simplify_iter_data(v)
                        File[k] = v
            else:
                File = {}
        else:
            raise TypeError(f'{f_ext} is not supported.')
        
        self.fpath = file_path
        self.ftype = f_ext[1:]
        self.fmode = mode
        self.File = File
    
    def _from_dict(
        self, file_path, group_dict, mode, allow_remove, plugins
    ):
        """Populates the data object from a dictionary.

        Parameters
        ----------
        file_path : :obj`str`
            Path to a file supported by reptar. If it does not exist, then one
            will be created if possible.
        group_dict : :obj:`str`
            Dictionary to populate the data object with.
        mode : :obj:`str`, optional
            A file mode string that defines the read/write behavior. Defaults to
            ``'r'``.
        allow_remove : :obj:`bool`, optional
            Allow the removal of exdir groups in ``w`` operation. Defaults to
            ``False``.
        plugins : :obj:`list`, optional
            A list of instantiated exdir plugins. Defaults to ``None``.
        """
        _, f_ext = os.path.splitext(file_path)
        
        if f_ext == '.exdir':
            self.File = exdir.File(
                file_path, mode, allow_remove, plugins
            )
        elif f_ext == '.json' or f_ext == '.npz':
            self.File = {}
        self.fpath = file_path
        self.ftype = f_ext[1:]
        self.fmode = mode

        for pair in self._iter_dict(group_dict):
            key = '/'.join(pair[:-1])
            data = pair[-1]
            self.add(key, data)
    
    def clean_key(self, key):
        """Clean key and remove any common mistakes.
        
        Parameters
        ----------
        key : :obj:`str`
            Key of the desired data.
        
        Returns
        -------
        :obj:`str`
            Cleaned key.
        """
        if '//' in key:
            key = key.replace('//', '/')
        return key
    
    def split_key(self, key):
        """Split the key into a parent and data key.

        Parameters
        ----------
        key : :obj:`str`
            Key of the desired data. Nested keys should be separated by ``/``.
        
        Returns
        -------
        :obj:`str`
            Parent key.
        :obj:`str`
            Data key.
        """
        if key[0] != '/':
            key = '/' + key
        key_split = key.rsplit('/', 1)
        if key_split[0] == '':
            key_split[0] = '/'
        return key_split
    
    def _get_from_dict(self, key):
        """Get data from dictionary-like file (e.g., json, npz).
        
        Parameters
        ----------
        key : :obj:`str`
            Key of the desired data. Nested keys should be separated by ``/``.
        """
        if key == '/':
            return self.File
        
        keys = key.split('/')
        if keys[0] == '':
            del keys[0]
        
        data = self.File[keys[0]]
        for k in keys[1:]:
            data = data[k]
        if isinstance(data, list):
            data_array = np.array(data)
            if data_array.dtype != 'O':
                data = data_array
        return data
    
    def _get_from_exdir(self, key):
        """Get data from exdir file.
        
        Parameters
        ----------
        key : :obj:`str`
            Key of the desired data. Nested keys should be separated by ``/``.
        """
        if key == '/':
            return self.File
        
        key_parent, key_data = self.split_key(key)
        parent = self.File[key_parent]
        if key_data in list(parent):
            data = parent[key_data]
            if isinstance(data, exdir.core.dataset.Dataset):
                data = data.data
        elif key_data in list(parent.attrs):
            data = parent.attrs[key_data]
            if isinstance(data, exdir.core.attribute.Attribute):
                data = dict(data)
        
        return data
    
    def get_keys(self, group_key):
        """A list of keys in a group.

        Does not include keys of nested groups.
        
        Parameters
        ----------
        group_key : :obj:`str`
            Key specifying which group to get the data keys from.
        
        Returns
        -------
        :obj:`list`
            A list of all available keys not leading to another group.
        """
        group_key = self.clean_key(group_key)
        group = self.get(group_key)

        keys = []
        if self.ftype == 'exdir':
            keys.extend(list(sorted(group.attrs.keys())))
            keys.extend(
                list(sorted(
                    key for key in group.keys() if not \
                    isinstance(group[key], exdir.core.raw.Raw)
                ))
            )
        elif self.ftype == 'json' or self.ftype == 'npz':
            keys.extend(list(group.keys()))
        return keys

    def get(self, key):
        """Retrieve data.

        Parameters
        ----------
        key : :obj:`str`
            Key of the desired data. Nested keys should be separated by ``/``.
        
        Examples
        --------
        >>> data.get('/prod_1/charge')
        0
        >>> data.get('energy_scf')
        -12419.360138637763
        """
        key = self.clean_key(key)
        if self.ftype == 'exdir':
            data = self._get_from_exdir(key)
        elif self.ftype == 'json' or self.ftype == 'npz':
            data = self._get_from_dict(key)
        return data
    
    def _is_iter(self, data):
        """If data is iterative (i.e., array, list, or tuple).
        
        Returns
        -------
        :obj:`bool`
        """
        if isinstance(data, np.ndarray) or isinstance(data, list) or isinstance(data, tuple):
            return True
        else:
            return False
    
    def simplify_iter_data(self, data):
        """Simplify iterative data if possible.

        Attempts to simplify iterative objects by the type and/or number of
        elements. For example, iterative data with only one element will return
        just that element.

        This is mainly used when adding data to a file. When parsing data we
        will not know if there will be multiple values of certain data (e.g.,
        multiple single point energies) until we finish parsing. Thus, we assume
        there will be multiple values while parsing then simplify any iterative
        objects at the end.
        
        Parameters
        ----------
        data : :obj:`numpy.ndarray`, :obj:`list`, :obj:`tuple`
            An iterative data object.
        """
        # If data is a list we check the types of data it contains.
        # If all of them are strings, we do not convert to array.
        if not isinstance(data, np.ndarray):
            # If there is only one item in the list we can likely
            # just store it as an attribute. However, we have
            # to be careful when the single item is an array.
            if len(data) == 1:
                data = data[0]
            # If all the items are not all strings then we make
            # the dataset.
            elif not all(isinstance(i, str) for i in data):
                data_array = np.array(data)
                # Sometimes we have a list of objects that cannot be easily
                # converted to numpy data types. Numpy will use an object
                # dtype, so we only convert to array if its not an 'O' dtype.
                if data_array.dtype != 'O':
                    data = data_array
            # At this point only data that contains all strings
            # should be left. We put these as attributes (i.e., 
            # comp_ids)
            else:
                pass
        # Handling arrays.
        else:
            # If only one item we store it as a value.
            if data.shape == (1,):
                data = data[0].item()
        
        return data
    
    def _add_to_dict(self, key, data):
        """Add data to dictionary-like file.

        Parameters
        ----------
        key : :obj:`str`
            Where to add the data. Can be a nested key.
        data : various
            Data to add to file.
        """
        if isinstance(data, np.ndarray):
            if 'U' in str(data.dtype):
                data = data.tolist()

        keys = key.split('/')
        if keys[0] == '':
            del keys[0]
        
        is_iter = self._is_iter(data)
        if is_iter:
            data = self.simplify_iter_data(data)

        add_dic = {keys[-1]: data}
        for key in reversed(keys[:-1]):
            add_dic = {key: add_dic}

        self.File = combine_dicts(self.File, add_dic)
    
    def _add_to_exdir(self, key, data):
        """Add data to an exdir group.

        Parameters
        ----------
        key : :obj:`str`
            Where to add the data. Can be a nested key.
        data : various
            Data to add to exdir file.
        """
        parent_key, data_key = self.split_key(key)
        group = self.get(parent_key)
        assert isinstance(group, exdir.core.exdir_file.File) \
            or isinstance(group, exdir.core.group.Group)
        
        # Custom handling of parsed data to ensure logical behavior.
        if data_key == 'dipole_moment':
            if isinstance(data, np.ndarray):
                data = data.tolist()
            group.attrs[data_key] = data
            return None
        elif data_key == 'wall_potential':
            group.attrs[data_key] = data
            return None

        # If data is not array/list or array/list of one dimension and
        # length we make it an exdir attribute.
        # We make anything else an exdir dataset.
        is_iter = self._is_iter(data)
        if is_iter:
            data = self.simplify_iter_data(data)
            if isinstance(data, np.ndarray):
                store_as = 'dset'
            else:
                store_as = 'attr'
        # Handles all non iterable data. 
        else:
            store_as = 'attr'
        
        if store_as == 'attr':
            group.attrs[data_key] = data
        elif store_as == 'dset':
            try:
                group.create_dataset(data_key, data=data)
            except RuntimeError:
                # If data set already exists, exdir will throw a RuntimeError.
                # We remove this dataset and try again.
                group.__delitem__(data_key)
                group.create_dataset(data_key, data=data)

    def add(self, key, data):
        """Add data to file.

        Note that there is some data postprocessing using
        :meth:`~reptar.data.data.simplify_iter_data`.

        Parameters
        ----------
        key : :obj:`str`
            Key of the desired data (including parent). Nested keys should be
            separated by ``/``.
        data
            Data to add to file.
        """
        key = self.clean_key(key)
        if self.ftype == 'exdir':
            self._add_to_exdir(key, data)
        elif self.ftype == 'json' or self.ftype == 'npz':
            self._add_to_dict(key, data)
    
    def _iter_dict(self, dic):
        """Iterate over nested dictionary.

        Parameters
        ----------
        dic : :obj:`dict`
            An arbitrarily nested dictionary.
        
        Yields
        ------
        A :obj:`tuple` of keys where the last element is the value.
        """
        for k, v in dic.items():
            if isinstance(v, dict):
                # If value is dict then iterate over all its values
                for pair in self._iter_dict(v):
                    yield (k, *pair)
            else:
                # If value is not dict type then yield the value
                yield (k, v)
    
    def init_group(self, key):
        """Initialize an exdir group with the specified key.

        Parameters
        ----------
        key : :obj:`str`
            Key of the desired data (including parent). Nested keys should be
            separated by ``/``.
        """
        parent_key, group_key = self.split_key(key)
        if self.ftype == 'exdir':
            parent = self.get(parent_key)
            group = parent.create_group(group_key)
        return group
    
    def as_dict(self, group_key):
        """Get a group as a dictionary.

        Parameters
        ----------
        group_key : :obj:`str`
            Desired group.
        
        Returns
        -------
        :obj:`dict`
            The desired group as a dictionary.
        """
        group = self.get(group_key)
        if self.ftype == 'json' or self.ftype == 'npz':
            return group
        elif self.ftype == 'exdir':
            group_dict = {}
            data_keys = self.get_keys(group_key)
            get_keys = [f'{group_key}/{d_key}' for d_key in data_keys]
            for d_key,g_key in zip(data_keys, get_keys):
                group_dict[d_key] = self.get(g_key)
            return group_dict
    
    def save(self, json_prettify=True):
        """Saves non-exdir files.

        Parameters
        ----------
        json_prettify : :obj:`bool`
            Indents JSON objects if True. If false the JSON file is only one
            line.
        """
        assert self.fmode == 'w'
        if self.ftype == 'json':
            json_dict = self.File

            if json_prettify:
                json_string = json.dumps(
                    json_dict, cls=cclib.io.cjsonwriter.JSONIndentEncoder,
                    sort_keys=True, indent=4
                )
            else:
                json_string = json.dumps(
                    json_dict, cls=cclib.io.cjsonwriter.NumpyAwareJSONEncoder,
                    sort_keys=True
                )
            with open(self.fpath, 'w') as f:
                f.write(json_string)
        elif self.ftype == 'npz':
            npz_dict = self.File
            np.savez_compressed(self.fpath, **npz_dict)



# MIT License
# 
# Copyright (c) 2022, Alex M. Maldonado
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from abc import ABC, abstractmethod

class parser(ABC):
    """Base class for parsing output files.

    Parameters
    ----------
    out_path : :obj:`str`
        Path to output file.
    extractors : :obj:`list`
        Additional extractors for the parser to use.
    
    Attributes
    ----------
    out_path : :obj:`str`
        Path to output file.
    file_name : :obj:`str`
        The name of the file without extension.
    """

    def __init__(self, out_path, extractors):
        self.out_path = out_path
        self.file_name = '.'.join(self.out_path.split('/')[-1].split('.')[:-1])
        self.parsed_info = {
            'system_info': {},
            'runtime_info': {},
            'outputs': {}
        }
        self.extractors = extractors
    
    @abstractmethod
    def parse(self):
        """Drive the parsing and postprocessing of parsed data.

        In general, there are usually three stages.

        :**Preprocessing**:
            Anything that is done before reptar does its parsing.
            Reptar does not have extractors for all information, so sometimes `cclib <https://cclib.github.io/>`_ is used to extract additional information.

        :**Extracting**:
            Iterates through all lines in :term:`out_path` using all extractors with :meth:`~reptar.parsers.parser.parser.extract_data_out`.
            Each extractor has its own ``parsed_info`` that needs to be added to :attr:`~reptar.parsers.parser.parser.parsed_info`.

            .. automethod:: reptar.parsers.parser.extract_data_out
                :noindex:

        :**Postprocessing**:
            Any additional tasks to complete parsing.
            There is usually a ``after_parse`` method that adds or changes data based on the final :attr:`~reptar.parsers.parser.parser.parsed_info` attribute.
            Handling of other files such as :term:`geom_path` and :term:`traj_path`.
        """
        return NotImplemented
    
    @property
    def parsed_info(self):
        """Information parsed from files. Contains the following keys.

        .. glossary::

            ``system_info``
                Information specifying the system prior to any computation. Such
                as the initial cartesian coordinates, total system charge and
                multiplicity, etc.
            
            ``runtime_info``
                Contains information about setting up the job/calculation or running
                the job. Defining convergence criteria, parameters, etc.
            
            ``outputs``
                Results, requested or not, of the job. For example, SCF
                cycle values, optimized coordinates, trajectory, number of
                electrons, generated structures, etc.

        :type: :obj:`dict`
        """
        return self._parsed_info
    
    @parsed_info.setter
    def parsed_info(self, value):
        self._parsed_info = value

    @parsed_info.deleter
    def parsed_info(self, value):
        del self._parsed_info
    
    def after_parse(self):
        """Replace if desired"""
        pass
    
    def extract_data_out(self):
        """Extract data from ``out_path`` using all extractors"""
        with open(self.out_path, mode='r') as f:
            for line in f:
                for extractor in self.extractors:
                    for i in range(len(extractor.triggers)):
                        if extractor.triggers[i][0](line):
                            getattr(extractor, extractor.triggers[i][1])(f, line)
                            break
                    else:
                        continue
                    break
        self.combine_extracted()
    
    def combine_extracted(self):
        """Combines all parsed_info from extractors into the parsed_info in the
        parser object.
        """
        for extractor in self.extractors:
            extracted_info = extractor.parsed_info
            for key_cat in extracted_info.keys():
                for key_def in extracted_info[key_cat].keys():
                    if key_def not in self.parsed_info[key_cat].keys():
                        self.parsed_info[key_cat][key_def] = extracted_info[key_cat][key_def]
                    else:
                        pass
    
    def map_cclib_data(self):
        """
        """
        parsed_info = self.parsed_info
        cclib_data = self.cclib_data

        # Loop through every attribute in cclib_data and adds data to parsed data.
        for attr in dir(cclib_data):
            if not attr.startswith('__'):
                if attr == 'atomcharges':
                    parsed_info['outputs']['charges_lowdin'] = \
                        cclib_data.atomcharges['lowdin']
                    parsed_info['outputs']['charges_mulliken'] = \
                        cclib_data.atomcharges['mulliken']

                if attr == 'atomcoords':
                    parsed_info['system_info']['geometry'] = \
                        cclib_data.atomcoords
                
                if attr == 'atomnos':
                    parsed_info['system_info']['atomic_numbers'] = \
                        cclib_data.atomnos

                if attr == 'charge':
                    parsed_info['system_info']['charge'] = \
                        int(cclib_data.charge)
                
                if attr == 'mult':
                    parsed_info['system_info']['mult'] = \
                        int(cclib_data.mult)
                
                if attr == 'grads':
                    parsed_info['outputs']['grads'] = \
                        cclib_data.grads
                
                if attr == 'moenergies':
                    parsed_info['outputs']['energy_mos'] = \
                        cclib_data.moenergies
                
                if attr == 'moments':
                    parsed_info['outputs']['dipole_moment'] = \
                        cclib_data.moments[1]
                
                if attr == 'scfenergies':
                    parsed_info['outputs']['energy_scf'] = \
                        cclib_data.scfenergies
                
                if attr == 'nbasis':
                    parsed_info['runtime_info']['basis_n_func'] = \
                        int(cclib_data.nbasis)
                
                if attr == 'nelectrons':
                    parsed_info['system_info']['n_ele'] = \
                        int(cclib_data.nelectrons)

        self.parsed_info = parsed_info
    



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages
import versioneer

requirements = [
    'numpy', 'exdir>=0.4.2', 'cclib>=1.7.0', 'scipy'
]

setup_requirements = ['versioneer']

test_requirements = requirements.append(['pytest', 'pytest-order'])

setup(
    install_requires=requirements,
    extras_require={},
    include_package_data=True,
    package_data={'': ['definitions/*.yaml']},
    packages=find_packages(include=['reptar', 'reptar.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    zip_safe=False,
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
)



# MIT License
# 
# Copyright (c) 2022, Alex M. Maldonado
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import numpy as np
from .parser import parser
from ..extractors import extractorXTB
from ..utils import atoms_by_number, parse_stringfile

class parserXTB(parser):
    """Custom parser for xtb output files.

    Parameters
    ----------
    out_path : :obj:`str`
        Path to the main log file generated by the package.
    geom_path : :obj:`str`, optional
        Path to a file containing a single geometry.
    traj_path : :obj:`str`, optional
        Path to a trajectory file from a geometry optimization, MD
        simulation, etc.
    extractors : :obj:`list`, optional
        Additional extractors for the parser to use.
    """

    def __init__(self, out_path, geom_path=None, traj_path=None, extractors=None):
        self.package = 'xtb'
        if extractors is None:
            extractors = []
        extractors.insert(0, extractorXTB())
        super().__init__(out_path, extractors)

        self.geom_path = geom_path
        self.traj_path = traj_path
        if traj_path is not None and geom_path is None:
            raise ValueError(
                'A geometry file must be specified with trajectory file'
            )
        
        self.parsed_info['runtime_info']['prov'] = 'xTB'
    
    def parse(self):
        """Parses trajectory file and extracts information.
        """
        # Extract information.
        self.extract_data_out()

        # Adding any structure information.
        Z = []
        comments = []
        R = []
        if self.geom_path is not None:
            Z_geom, _, R_geom = parse_stringfile(self.geom_path)
            Z.extend(Z_geom)
            R.extend(R_geom)
        if self.traj_path is not None:
            Z_traj, comments, R_traj = parse_stringfile(self.traj_path)
            Z.extend(Z_traj)
            R.extend(R_traj)

            energy_pot = []
            # grad_norm = []
            for comment in comments:
                comment = comment.split()
                energy_pot.append(float(comment[1]))
                # grad_norm.append(float(comment[3]))
            self.parsed_info['outputs']['energy_pot'].extend(
                energy_pot
            )
        
        if len(set(tuple(i) for i in Z)) == 1:
            Z = Z[0]
        else:
            raise ValueError('Atomic numbers are not consistent.')
        Z = np.array(atoms_by_number(Z))
        R = np.array(R)
        if R.ndim == 2:
            R = np.array([R])
        assert R.ndim == 3

        self.parsed_info['system_info']['atomic_numbers'] = Z
        self.parsed_info['system_info']['geometry'] = R

        self.after_parse()
        return self.parsed_info
    

    def after_parse(self):
        """Checks to perform after parsing output file.
        """
        # xtb prints the last energy twice during optimizations.
        # The last printed energy has more significant figures, so we will
        # get rid of the second to last one.
        # It is also unclear why the structure considered "CYCLE   1" is
        # missing from the geometry log. It goes from the provided structure
        # to the "CYCLE   2" structure. So we will also remove this energy.
        # Note that we overwrite this data later, but it is better to ensure
        # consistency than punting to later in the code.
        if self.parsed_info['runtime_info']['calc_driver'] == 'optimization':
            del self.parsed_info['properties']['energy_scf'][1]
            del self.parsed_info['properties']['energy_scf'][-2]
        
        if 'success' not in self.parsed_info['runtime_info'].keys():
            self.parsed_info['runtime_info']['success'] = False
    
    
    
    
    
   
